/* sockcat Copyright (c) 2002 Michael Wolf.  All rights reserved.  */

/* usage: 
   sockcat -p PATH -i IP -t PORT -y stream|datagram|seqpacket -[c|l] -[h]  

PATH is the path of a local socket;  IP is an IP address to connect to or
listen on; -t is a port to connect to or bind; -c tells sockcat to connect; 
-l tells sockcat to listen.  -h prints help and exits. 

Using local and ip sockets are, naturally, mutually exclusive.  */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>

#include "socket.h"
#include "byte_array.h"
#include "util.h"

#include "xmalloc.h"
#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

#define SUN_PATH_MAX 104

#define STDIN_READ_SIZE 64

enum action { BOGUS = 0, LISTEN, CONNECT };

enum socket_domain domain;
enum socket_type type;
char *path;
u16 port;
u32 ip;
enum action action = BOGUS;

/* These are global so the signal handler can access them.  */
socket_t *sock;
socket_t *p;

void process_args(int, char **);
void validate(void);
socket_address_t *fill_address(void);
socket_t *obtain_working_socket(socket_t *, enum action);
void usage(void);
void sighandler(int);

int
main(int argc, char **argv)
{
    int ret;
    socket_address_t *address;
    struct timeval timeout;
    fd_set rset, wset;
    size_t size;
    byte_array *data;
    byte_array *idata;
    
    signal(SIGINT, sighandler);
    signal(SIGPIPE, sighandler);
    signal(SIGTERM, sighandler);

    sock = NULL;
    p = NULL;

    process_args(argc, argv);
    validate();
    address = fill_address();

    if (action == LISTEN) {
        sock = socket_listener(address, domain, type);
    } else {
        sock = socket_sender(address, domain, type);
    }
    if (sock == NULL) {
        die("unable to create socket\n");
    }
    if (domain == SOCKET_LOCAL)
        xfree(address->local.path);
    xfree(address);

    p = obtain_working_socket(sock, action);
    

    idata = (byte_array *) xcalloc(1, sizeof(byte_array));
    idata->data = (char *) xcalloc(STDIN_READ_SIZE, sizeof(char));
    while (1) {
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        FD_SET(p->fd, &rset);
        FD_SET(fileno(stdin), &rset);
        // FD_SET(p->fd, &wset);
        timeout.tv_sec = 10;
        timeout.tv_usec = 0;

        ret = select(p->fd + 1, &rset, &wset, NULL, &timeout);
        if (ret == -1)
            die("select failed");
        if (ret == 0) {
            printf_on_debug("select returned 0\n");
            continue;
        }

        if (FD_ISSET(p->fd, &rset)) {
            data = (byte_array *) xcalloc(1, sizeof(byte_array));
            /* 64 is just some randomish number.  It really only has
               an effect with SOCKET_STREAM anyway.  */
            size = socket_recv(p, data, 64); 
            printf_on_debug("read %d bytes from socket\n", size);
            if (size == 0) {
                /* This means the remote end has closed the socket.  */
                byte_array_destroy(data);
                goto done;
            }
            write(fileno(stdout), data->data, data->len);
            byte_array_destroy(data);
        }
        if (FD_ISSET(fileno(stdin), &rset)) {
            ret = read(fileno(stdin), idata->data, STDIN_READ_SIZE - 1);
            idata->len = ret;
            // printf_on_debug("DEBUG: read '%s'\n", data->data);
            printf_on_debug("read %d bytes from stdin\n", ret);
            // printf("%s", data->data);
            if (ret == 0)
                break;
            socket_send(p, idata);
    
        }
    }
done:
    byte_array_destroy(idata);
    socket_close(p, SOCKET_BOTH);
    if (action == LISTEN)
        socket_close(sock, SOCKET_BOTH);
    // return 0;
#ifdef WITH_DMALLOC
    dmalloc_shutdown();
#endif
    exit(0);
}

socket_address_t *
fill_address()
{
    socket_address_t *retval;

    retval = (socket_address_t *) xcalloc(1, sizeof(*retval));
    if (domain == SOCKET_INTERNET) {
        retval->internet.port = port;
        retval->internet.ip = ip;
    } else if (domain == SOCKET_LOCAL) {
        retval->local.path = (char *) xcalloc(SUN_PATH_MAX + 1, sizeof(char));
        memset(retval->local.path, '\0', SUN_PATH_MAX);
        strncpy(retval->local.path, path, SUN_PATH_MAX - 1);
    } else {
        croak("faulty argument parsing\n");
    }
    return retval;
}

void
process_args(int argc, char **argv)
{
    int c;
    int ret;
    struct in_addr ia;

    path = NULL;
    port = 0;
    ip = INADDR_NONE;
    
    while ((c = getopt(argc, argv, "p:i:t:y:lch")) != -1) {
        switch (c) {
        case 'p':
            path = optarg;
            break;
        case 'i':
            ret = inet_aton(optarg, &ia);
            if (ret == 0)
                die("bogus ip address");
            ip = ia.s_addr;
            break;
        case 't':
            port = atoi(optarg);
            if (port <= 0 || port > 65536)
                die("bogus port");
            break;
        case 'y':
            if (strcmp("stream", optarg) == 0)
                type = SOCKET_STREAM;
            else if (strcmp("datagram", optarg) == 0)
                type = SOCKET_DATAGRAM;
            else if (strcmp("seqpacket", optarg) == 0)
                type = SOCKET_SEQPACKET;
            else
                die("unknown type\n");
            break;
        case 'l':
            action = LISTEN;
            break;
        case 'c':
            action = CONNECT;
            break;
        case 'h':
            usage();
            exit(0);
            break;
        default:
            fprintf(stderr, "unknown thingy");
            break;
        }
    }
}    

void
validate()
{
    if (path != NULL && port != 0) 
        die("can't use a port number and a local socket");
    if (path != NULL && ip != INADDR_NONE)
        die("can't use an ip address and a local socket");
    if (path != NULL)
        domain = SOCKET_LOCAL;
    else
        domain = SOCKET_INTERNET;
    if (action == BOGUS)
        die("need an action type\n");
}

void
usage()
{
    printf("usage: 
   sockcat -p PATH -i IP -t PORT -y stream|datagram|seqpacket -[c|l] -[h]  

PATH is the path of a local socket;  IP is an IP address to connect to or
listen on; -t is a port to connect to or bind; -c tells sockcat to connect; 
-l tells sockcat to listen.  -h prints help and exits. 

Using local and ip sockets are, naturally, mutually exclusive.\n");
    exit(0);


}

socket_t *
obtain_working_socket(socket_t *csock, enum action act)
{
    socket_t *retval;
    int ret;
    fd_set rset;

    if (act == LISTEN) {
        FD_ZERO(&rset);
        FD_SET(csock->fd, &rset);
        ret = select(csock->fd + 1, &rset, NULL, NULL, NULL);
        if (ret == -1) 
            croak("select failed: %s", strerror(errno));
        retval = socket_accept(csock);
        if (retval == NULL)
            die("socket_accept failed");
        printf_on_debug("accepted connection\n");
    } else {
        retval = csock;
    }
    return retval;
}

void
sighandler(int sig)
{
    int ret;
    ret = 0;
    switch (sig) {
    case SIGTERM:
        printf_on_debug("DEBUG: caught SIGTERM\n");
        if (p != NULL)
            ret += socket_close(p, SOCKET_BOTH);
        if (sock != NULL)
            ret += socket_close(sock, SOCKET_BOTH);
        break;
    case SIGINT:
        printf_on_debug("DEBUG: caught SIGINT\n");
        if (p != NULL)
            ret += socket_close(p, SOCKET_BOTH);
        if (sock != NULL)
            ret += socket_close(sock, SOCKET_BOTH);
        break;
    case SIGPIPE:
        printf_on_debug("DEBUG: caught SIGPIPE (uh-oh)\n");
        if (p != NULL)
            ret += socket_close(p, SOCKET_BOTH);
        if (sock != NULL)
            ret += socket_close(sock, SOCKET_BOTH);
        break;
    }
    if (ret != 2) {
        printf("WARNING: there may have been problems cleaning up\n");
        exit(1);
    }
    exit(0);
}

