/* Copyright (c) 2002 Omar Kilani.  All rights reserved.
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * moodns is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Header: /cvsroot/moodns/moodns/src/byte_array.c,v 1.7 2002/07/20 16:54:59 mawolf Exp $  */

#ifndef _REENTRANT
#define _REENTRANT
#endif

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "byte_array.h"

#include "xmalloc.h"
#ifdef WITH_DMALLOC
#  include <dmalloc.h>
#endif

static u8 *
byte_copy_data(const u8 *src, size_t len)
{
   u8 *dst;

   dst = (u8 *) xmalloc(len);
   if (dst != NULL) {
       memcpy(dst, src, len); 
       return dst;
   } else {
       return NULL;
   }
} 

size_t
byte_array_cat(byte_array * orig, byte_array * tail)
{
    assert(orig != NULL);
    assert(tail != NULL);
    assert(tail->data != NULL);
    assert(tail->len != 0);

    return byte_array_cats(orig, tail->data, tail->len);
}

size_t
byte_array_cats(byte_array * orig, const u8 * data, size_t len)
{
    u8 *x;
    u8 *y = byte_copy_data(data, len); 
    
    assert(orig != NULL);
    assert(data != NULL);
    assert(len != 0);
    
    x = (u8 *) xrealloc(orig->data, orig->len + len);
    if (x != NULL) {
        orig->data = x;
    } else {
        xfree(y);
        return 0;
    }
    memcpy(orig->data + orig->len, y, len);
    orig->len += len;
    xfree(y);
    return orig->len;
}

size_t
byte_array_catu8(byte_array *orig, u8 i)
{
    return byte_array_cats(orig, (u8 *) &i, sizeof(u8));
}

size_t
byte_array_catu16(byte_array *orig, u16 i)
{
    return byte_array_cats(orig, (u8 *) &i, sizeof(u16));
}

size_t
byte_array_catu32(byte_array *orig, u32 i)
{
    return byte_array_cats(orig, (u8 *) &i, sizeof(u32));
}

size_t
byte_array_insert(byte_array * orig, size_t idx, byte_array * insert)
{
   assert(orig != NULL);
   assert(insert != NULL);
   assert(insert->data != NULL);
   assert(insert->len != 0);

   return byte_array_inserts(orig, idx, insert->data, insert->len);
}

size_t
byte_array_inserts(byte_array * orig, size_t idx, u8 * data, size_t len)
{
   u8 *x;
   u8 *y = byte_copy_data(data, len);

   assert(orig != NULL);
   assert(data != NULL);
   assert(len != 0);

   x = (u8 *) realloc(orig->data, orig->len + len);
   if (x != NULL) {
       orig->data = x;
   } else {
       xfree(y);
       return 0;
   }
   memmove(orig->data + idx + len,  orig->data + idx, orig->len - idx); 
   memcpy(orig->data + idx, y, len);
   orig->len += len;
   xfree(y);
   return orig->len; 
} 

size_t
byte_array_delete(byte_array * orig, size_t idx, size_t len)
{
   u8 *x;

   assert(orig != NULL);
   assert(idx <= orig->len);
   assert(len != 0);

   memmove(orig->data + idx, orig->data + idx + len, orig->len - idx); 
   x = (u8 *) realloc(orig->data, orig->len - len);

   if (x != NULL)
      orig->data = x;

   orig->len -= len;
   return orig->len;
}

size_t
byte_array_replace(byte_array * orig, size_t idx, byte_array * replace)
{
   assert(orig != NULL);
   assert(idx <= orig->len);
   assert(replace != NULL);
   assert(replace->data != NULL);
   assert(replace->len != 0);

   return byte_array_replaces(orig, idx, replace->data, replace->len);
}

size_t
byte_array_replaces(byte_array * orig, size_t idx, u8 * data, size_t len)
{
   u8 *y = byte_copy_data(data, len);

   assert(orig != NULL);
   assert(idx <= orig->len);
   assert(data != NULL);
   assert(len != 0);

   if (idx + len > orig->len) {
       xfree(y);
       return 0; 
   }
   memcpy(orig->data + idx, y, len);  
   xfree(y);
   return orig->len;
}

void
byte_array_destroy(byte_array * orig)
{
    /* printf("orig = %p, orig->data = %p\n", orig, orig->data);  */
    if (orig != NULL) {
        if (orig->data != NULL) {
            xfree(orig->data);
            orig->data = NULL;
        }
        xfree(orig);
        orig = NULL;
    }
}

#if TEST
int main(int argc __attribute__((unused)), char **argv __attribute__((unused)))
{
    byte_array *t = (byte_array *) xcalloc(1, sizeof(byte_array));
    byte_array *ins = (byte_array *) xcalloc(1, sizeof(byte_array));
    byte_array *rep = (byte_array *) xcalloc(1, sizeof(byte_array));
    char buf[1024];

    sprintf(buf, "insert"); 
    byte_array_cats(ins, buf, strlen(buf));
    sprintf(buf, "hello"); 
    byte_array_cats(t, buf, strlen(buf));
    printf("data = %-*.*s, len = %d\n", t->len, t->len, t->data, t->len);
    printf("data = %-*.*s, len = %d\n", ins->len, ins->len, ins->data, ins->len);
    printf("data = %-*.*s, len = %d\n", t->len, t->len, t->data, t->len);
    sprintf(buf, "replace");
    byte_array_cats(rep, buf, strlen(buf));
    printf("data = %-*.*s, len = %d\n", rep->len, rep->len, rep->data, rep->len);
    byte_array_replace(t, 3, rep); 
    printf("data = %-*.*s, len = %d\n", t->len, t->len, t->data, t->len);
    byte_array_inserts(t, 1, "i ", 2);
    printf("data = %-*.*s, len = %d\n", t->len, t->len, t->data, t->len);
    byte_array_replaces(t, 2, "m", 1); 
    printf("data = %-*.*s, len = %d\n", t->len, t->len, t->data, t->len);
    byte_array_delete(t, 2, 1); 
    printf("data = %-*.*s, len = %d\n", t->len, t->len, t->data, t->len);

    byte_array_destroy(t);
    byte_array_destroy(ins);
    byte_array_destroy(rep);
    return 0;
}
#endif
