/* Copyright (c) 2001, 2002 Michael Wolf.  All rights reserved. 
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * moodns is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
    
/* $Header: /cvsroot/moodns/moodns/src/cache.c,v 1.7 2002/03/05 01:59:34 mawolf Exp $  */

/* Here's a big picture description of the cache.

   
   External programmer's interface:
   
   A client searches for a record via cache_check.  A record chain
   containing all (cached) records of the appropriate type is
   returned.  (It is up to a client to free that record_chain and
   the records contained therein because a deep copy is made.  Stupid
   C; this crap wouldn't be necessary if we were using garbage
   collection.) 
   
   A client can add a record via cache_put.  A client can add a
   record_chain via cache_put_record_chain.  These functions are
   asynchronous, so records added there might not be immediately
   available to other clients.  Thus, a client adding these records
   should also use and deallocate them itself.  (More of the deep 
   copying crap.)  Redundant writes to the cache aren't a problem.  
   
   
   Implementation details:

   Data is stored in two structures.

   The first structure is a ht storing `record_chain's.  The key used
   is a domain name.  

   The second structure is a heap of `record's.  These are stored by
   ttl.  

   Writes to the cache, as mentioned above, are asynchronous.  This is
   done via two socketpairs.  (We use socketpairs rather than pipes
   because pipes have limited buffer space which could be overrun, 
   although this is unlikely.)  Rather than sending all the data
   across the sockets, only pointers to the data are sent over the
   sockets.   

   There are two independent threads associated with the cache.  The
   first is adder_thread.  The second is trimmer_thread.  
   
   adder_thread sits in a loop.  It calls select on both of the sockets,
   and reads pointers from them as they're available.  It locks the
   cache and adds the data.  (NB: maybe it could wait a while before
   doing this so that the cache isn't being write-locked too often.)

   trimmer_thread sits in a loop.  When it decides that the cache has
   grown too large, it deletes the oldest entries.  When it decides
   that the oldest entries are too old, it deletes them too.  


   XXX: We need to use either deep copies or some form of automatic
   reclamation here, because everything else will be prone to race
   conditions.  This is necessary with both additions to and
   extractions from the cache.  (I wonder if this has been documented
   by the "patterns community", or in other literature that discusses
   implementation of caches in general.)  


   It would be useful to create a single-threaded version of the cache
   as well, for use by st-test and whatever else.  Plus, this need for
   deep copies would go away, although it'll stay because interface
   consistency is important.  

   XXX: Factor out common code between threaded and non-threaded
   versions.  
*/

/* Known issues:

   - It doesn't work yet.  The whole record / record_chain thing is a bit
   dodgy still.  Need to work on that.  

   - Modulo the record / record_chain issues, I'm still not sure if the
   cache really works yet.  I think it does, but I haven't gotten to properly 
   testing it yet.   

   - I think we're inserting putting the wrong thing into the table and the 
   heap.  We should be adding to `record_chain's instead.

   - Is having a thread inspect and limit the cache really the best way to 
   keep it limited in size to avoid DOS style attacks?  Maybe there should be 
   "soft" and "hard" limits.  
   
*/

#ifndef _REENTRANT
#define _REENTRANT
#endif

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <pthread.h>

#include "wpthread.h"

#include "record.h"
#include "heap.h"
#include "ht.h"
#include "mtime.h"
#include "cache.h"

#include "xmalloc.h"

#if WITH_DMALLOC
#include <dmalloc.h>
#endif

struct cache {
    int threaded;
    
    pthread_mutex_t *mutex;

    pthread_cond_t *reader_cv;
    pthread_cond_t *writer_cv;

    heap *ttls;
    ht *table;

    /* -1 if we don't care  */
    int max_size;
    int current_size;

    /* 0 if nobody is doing anything.  > 0 if there are readers.  -1 if there
       is a writer.  There may only be one writer.  */
    int rw_status;
    
    int new_readers_allowed;


    int record_read;
    int record_write;
    pthread_mutex_t *record_write_mutex;

    int record_chain_read;
    int record_chain_write;
    pthread_mutex_t *record_chain_write_mutex;
    
    pthread_t *adder_thread;
    pthread_t *trimmer_thread;
};

static void *_trimmer(void *);
static void *_adder(void *);
static int _record_compare_local(void *, void *, void *);
static int _record_ttl_compare_local(void *, void *, void *);
static void _record_delete_local(void *, void *);
static void _destructor(cache *);

cache *
cache_init()
{
    cache *retval;
    int sp1[2], sp2[2];
    int ret;
    int error = 0;

    retval = (cache *) xcalloc(1, sizeof(cache));

    retval->threaded = 1;

    retval->mutex = (pthread_mutex_t *) xmalloc(sizeof(pthread_mutex_t));
    wpthread_mutex_init(retval->mutex, NULL);

    retval->reader_cv = (pthread_cond_t *) xmalloc(sizeof(pthread_cond_t));
    wpthread_cond_init(retval->reader_cv, NULL);

    retval->writer_cv = (pthread_cond_t *) xmalloc(sizeof(pthread_cond_t));
    wpthread_cond_init(retval->writer_cv, NULL);

    retval->ttls = heap_init(_record_ttl_compare_local);

    retval->table = ht_init(_record_compare_local, _record_delete_local);

    retval->max_size = DEFAULT_MAX_SIZE; /* yeeuch  */
    retval->current_size = 0;

    retval->adder_thread = (pthread_t *) xmalloc(sizeof(pthread_t));
    retval->trimmer_thread = (pthread_t *) xmalloc(sizeof(pthread_t));

    ret = socketpair(PF_LOCAL, SOCK_STREAM, 0, sp1);
    if (ret == -1) {
        perror("socketpair failed");
        error = 1;
    }
    retval->record_read = sp1[0];
    retval->record_write = sp1[1];

    ret = socketpair(PF_LOCAL, SOCK_STREAM, 0, sp2);
    if (ret == -1) {
        perror("socketpair failed");
        error = 1;
    }
    
    retval->record_chain_read = sp2[0];
    retval->record_chain_write = sp2[1];


    /* These MUST be the last thing before the return statement.  */
    wpthread_create(retval->adder_thread, NULL, _adder, retval);
    wpthread_create(retval->trimmer_thread, NULL, _trimmer, retval);

    if (error) {
        _destructor(retval);
        retval = NULL;
    }

    return (retval);
}

cache *
cache_init_st()
{
    cache *retval;

    retval = xcalloc(1, sizeof(cache));
    
    retval->ttls = heap_init(_record_ttl_compare_local);

    retval->table = ht_init(_record_compare_local, _record_delete_local);
    
    retval->max_size = DEFAULT_MAX_SIZE; /* yeeuch  */
    retval->current_size = 0;

    return retval;
}


/* XXX: What happens if the cache trims itself?  It someone else has
   data next to be destroyed, that could be bad.  Solution seems to be
   a deep copy.  Yet another reason to use a gc.  Maybe reference
   counts will be sufficient in this case?  That would mean an extra
   lock per record structure.  Hmm.  No easy answers.  */
struct record_chain *
cache_check(cache *c, char *name, enum record_type type)
{
    struct record_chain *retval = NULL;
    struct record_chain *tmp;
    record *rec;
    int i;
    // record *tmp;

    if (!c->threaded)
        return cache_check_st(c, name, type);
       

    wpthread_mutex_lock(c->mutex);
    while (c->new_readers_allowed == -1) { /* Was 0, wrongly I believe.  */
        wpthread_cond_wait(c->reader_cv, c->mutex);
    }

    c->rw_status++;
    tmp = (struct record_chain *) ht_get(c->table, name);
    c->rw_status--;

    if (!tmp)
        goto done;
    
    retval = record_chain_init(name);
    for (i = 0; i < tmp->records_len; ++i) {
        if (tmp->records[i].type == type) {
            rec = (record *) xmalloc(sizeof(record));
            memmove(rec, &tmp->records[i], sizeof(*rec));
            record_chain_add(retval, rec);
        }
    }

done:
    if (c->new_readers_allowed != 0)
        wpthread_cond_signal(c->reader_cv);
    else
        wpthread_cond_signal(c->writer_cv);
    
    wpthread_mutex_unlock(c->mutex);

    return retval;
}

struct record_chain *
cache_check_st(cache *c, char *name, enum record_type type)
{
    struct record_chain *retval = NULL;
    struct record_chain *tmp;
    record *rec;
    int i;
    
    tmp = ht_get(c->table, name);
    
    if (!tmp)
        goto done;
    
    for (i = 0; i < tmp->records_len; ++i) {
        if (tmp->records[i].type == type) {
            rec = (record *) xmalloc(sizeof(record));
            memmove(rec, &tmp->records[i], sizeof(*rec));
            record_chain_add(retval, rec);
        }
    }
done:
    return retval;
}


int
cache_put(cache *c, record *r)
{
    int retval = 0;
    record *rec;
    int ret;

    if (!c->threaded)
        return cache_put_st(c, r);

    rec = record_alloc_dup(r);    
    wpthread_mutex_lock(c->record_write_mutex);
        
    ret = send(c->record_write, rec, sizeof(rec), 0);
    if (ret != sizeof(rec)) 
        abort();

    wpthread_mutex_unlock(c->record_write_mutex);

    return retval;
}


int
cache_put_st(cache *c, record *r)
{
    int retval = 0;
    record *rec;
    record *rt;
    unsigned int present;

    rec = record_alloc_dup(r);
    rec->ttl += now();
    heap_put(c->ttls, r, NULL);
    ht_put(c->table, strdup(r->name), r, NULL);
    ++c->current_size;

    /* Now check for entries that should be discarded.  */
    present = now();
    while (1) {
        rt = heap_peek(c->ttls);
        if (rt == NULL)
            break;
        if (record_expiry(rt) < present)
            break;
        rt = heap_get(c->ttls, NULL);
        ht_remove(c->table, rt->name);
        fprintf(stdout, "DEBUG: removing expired entry \"%s\" from cache.\n",
                rt->name);
        record_delete(rt);
        --c->current_size;
    }
    return retval;
}



int
cache_put_record_chain(cache *c __attribute__((unused)), 
                       struct record_chain *rc __attribute__((unused)))
{
    UNIMPLEMENTED;
    return 0;
#if 0
    struct record_chain *trc;
    record *tr;
    int i;
    int ret;

    if (!c->threaded)
        return cache_put_record_chain_st(c, rc);

    trc = record_chain_init(rc->name);
    for (i = 0; i < rc->records_len; ++i) {
        tr = record_alloc_dup(&rc->records[i]);
        record_chain_add(trc, tr); 
    }
    
    wpthread_mutex_lock(c->record_chain_write_mutex);

    ret = send(c->record_chain_write, trc, sizeof(trc), 0);
    if (ret != sizeof(trc))
        /* If this happens, I think we're screwed.  */
        abort();

    wpthread_mutex_unlock(c->record_chain_write_mutex);

    return 0;
#endif
}

int
cache_put_record_chain_st(cache *c __attribute__((unused)),
                          record_chain *rc __attribute__((unused)))
{
    UNIMPLEMENTED;
    return 0;
#if 0
    int retval;
    record *rec;
    record *rt;
    unsigned int present;
    int i;

    for (i = 0; i < rc->records_len; ++i) {
        rec = record_alloc_dup(&rc->records[i]);
        rec->ttl += now();
        heap_put(c->ttls, rec, NULL);
        ht_put(c->table, strdup(rec->name), rec, NULL);
        ++c->current_size;
    }

    present = now();
    while (1) {
        rt = heap_peek(c->ttls);
        if (rt == NULL)
            break;
        if (record_expiry(rt) < present)
            break;
        rt = heap_get(c->ttls, NULL);
        ht_remove(c->table, rt->name);
        fprintf(stdout, "DEBUG: removed expired entry \"%s\" from cache.\n",
                rt->name);
        record_delete(rt);
        --c->current_size;
    }
    return retval = 0;
#endif
}



#define MAX(a, b) (a) > (b) ? (a) : (b)

void *
_adder(void *_arg)
{
    cache *c = (cache *) _arg;
    fd_set *set = (fd_set *) xmalloc(sizeof(*set));
    int mfd = MAX(c->record_read, c->record_chain_read);
    struct timeval *tv = (struct timeval *) xmalloc(sizeof(struct timeval));
    int ret;

    printf("DEBUG: starting the cache async-additions thread.\n");

    while (1) {
        FD_ZERO(set);
        FD_SET(c->record_read, set);
        FD_SET(c->record_chain_read, set);
        tv->tv_sec = 60;
        tv->tv_usec = 0;
        
        ret = select(mfd + 1, set, NULL, NULL, tv);
        if (ret == 0) {
            printf("select timed out.  but that's ok.\n");
            continue;
        } else if (ret == -1) {
            perror("select failed");
            printf("We're probably screwed, but we'll keep trying.\n");
            continue;
        } 
        wpthread_mutex_lock(c->mutex);
        c->new_readers_allowed = 0;
        while (c->rw_status > 0) {
            wpthread_cond_wait(c->writer_cv, c->mutex);
        }
        if (FD_ISSET(c->record_read, set)) {
            record *r;
            int rv;
            rv = recv(c->record_read, &r, sizeof(r), 0);
            if (rv != sizeof(r))
                abort();
            printf("DEBUG: putting in %s\n", r->name);
            heap_put(c->ttls, r, NULL);
            ht_put(c->table, strdup(r->name), r, NULL);
            ++c->current_size;
            /* Hmm...  */    
        } else if (FD_ISSET(c->record_chain_read, set)) {
            struct record_chain *rc;
            int rv;
            rv = recv(c->record_chain_read, &rc, sizeof(rc), 0);
            if (rv != sizeof(rc))
                abort();
            /* iterate through.  slap them in.  how exciting.  */
        } else {
            abort();
        }
        c->new_readers_allowed = 1;
        // ++c->current_size;
        wpthread_cond_signal(c->reader_cv);
        wpthread_cond_signal(c->writer_cv);
        wpthread_mutex_unlock(c->mutex);
    }
    
    return NULL;
}


/* What does this thread do?  

   Periodically, check the state of the cache.  There are two tasks:
   a) Ensure that there are no records in the cache whose TTL has expired.
   b) Ensure that the cache isn't too large.  This is optional.

   0. Goto 2.

   1. Sleep a while.  

   2. Repeatedly examine the expiry of the oldest record (heap_peek) and remove
   if necessary.  Remove from the hash table too.

   3. Check the size of the cache against the expected max size.  If it's too
   large, keep taking records out of the heap and table until we're right.

   4. Goto 1.  

   This needs to be thread-safe (seems to be ok) and statistics should be 
   periodically logged.  What gets logged ought to be configurable.  */   
void *
_trimmer(void *_arg)
{
    cache *arg = (cache *) _arg;
    struct timeval tv;
    record *r;
    unsigned int present;

    fprintf(stdout, "DEBUG: starting the cache cleanup thread\n");

    while (1) {
        sleep(600);
        wpthread_mutex_lock(arg->mutex);
        fprintf(stdout, "DEBUG: cleaning the cache\n");
        fprintf(stdout, "DEBUG: there are %d things in the cache\n", 
                arg->current_size);
        
        arg->new_readers_allowed = 0;

        while (arg->rw_status > 0) {
            wpthread_cond_wait(arg->writer_cv, arg->mutex);
        }

        gettimeofday(&tv, NULL);
        present = tv.tv_sec;
        while (1) {
            r = heap_peek(arg->ttls);
            if (r == NULL)
                break;
            if (record_expiry(r) < present)
                break;
            r = heap_get(arg->ttls, NULL);
            ht_remove(arg->table, r->name);
            fprintf(stdout, "DEBUG: removing expired record \"%s\" from cache.\n", r->name);
            record_delete(r);
            --arg->current_size;
        }
        while (arg->current_size > arg->max_size) {
            r = heap_get(arg->ttls, NULL);
            ht_remove(arg->table, r->name);
            record_delete(r);
            --arg->current_size;
        }

        arg->new_readers_allowed = 1;
        wpthread_cond_signal(arg->reader_cv);
        wpthread_cond_signal(arg->writer_cv);
        wpthread_mutex_unlock(arg->mutex);
    }

    return NULL;
}

void
_destructor(cache *c)
{
    /* XXX: This function probably isn't entirely correct.  I want
       garbage collection!  */
    if (c->mutex)
        xfree(c->mutex);
    if (c->reader_cv)
        xfree(c->reader_cv);
    if (c->writer_cv)
        xfree(c->writer_cv);
    if (c->ttls)
        heap_destroy(c->ttls, NULL, NULL);
    if (c->table)
        ht_destroy(c->table, NULL, NULL);
    if (c->record_write_mutex)
        xfree(c->record_write_mutex);
    if (c->record_chain_write_mutex)
        xfree(c->record_chain_write_mutex);
    if (c->adder_thread)
        xfree(c->adder_thread); /* dodgy  */
    if (c->trimmer_thread)
        xfree(c->trimmer_thread); /* dodgy too  */
}



int
_record_compare_local(void *a, void *b, void *aux __attribute__((unused)))
{
    return (record_compare((record *) a, (record *) b));
}

int
_record_ttl_compare_local(void *a, void *b, void *aux __attribute__((unused)))
{
    return (record_ttl_compare((record *) a, (record *) b));
}

void
_record_delete_local(void *a, void *aux __attribute__((unused)))
{
    record_delete((record *) a);
}
