/*
 * Simple, robust heap / priority queue implementation.  
 *
 * Copyright (c) 2001 Michael Wolf (mike@dodgysoftware.net)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* $Header: /cvsroot/moodns/moodns/src/heap.c,v 1.1.1.1 2001/12/21 02:42:59 mawolf Exp $  */

/* Based partly on algorithms presented in Robert Sedgewick's _Algorithms
   in C_, 3rd ed, but be sure to see the comments prior to fix_up and 
   fix_down  */

#if HAVE_CONFIG_H
  #include "../config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "xmalloc.h"

#include "heap.h"

#if WITH_DMALLOC
  #include <dmalloc.h>
#endif


/* Some power of 2, minus 1.  */
static const unsigned int SIZE = 255; /*= 15; */ /* 255  */

static void _heap_iterate_gen(heap *, void (*)(void *, void *), void *);
static void _heap_debug_dump_recursive(heap *, void (*)(void *), int, char *);

static void _fix_up(heap *, void *);
static void _fix_down(heap *, void *);
static void _fix_down_recursive(heap *, void *, int);

struct heap {
    /* The number of elements.  */
    int size;

    /* The number of elements used.  */
    int used;
    /* The data stored.  */
    void **data;
    /* Comparison function.  */
    int (*cmp)(void *a, void *b, void *c);

    void (*get_hook)(void *data, void *aux);
    void *get_hook_aux;

    void (*put_hook)(void *data, void *aux);
    void *put_hook_aux;

    void (*peek_hook)(void *data, void *aux);
    void *peek_hook_aux;
};

heap *
heap_init(int (*compare)(void *, void *, void *))
{
    heap *retval;

    retval = (heap *) xcalloc(1, sizeof(heap));
    retval->size = SIZE;
    retval->used = 0;
    retval->data = xcalloc(retval->size, sizeof(void *));
    retval->cmp = compare;
    retval->get_hook = NULL;
    retval->get_hook_aux = NULL;
    retval->put_hook = NULL;
    retval->put_hook_aux = NULL;
    retval->peek_hook = NULL;
    retval->peek_hook_aux = NULL;

    return (retval);
}

int
heap_put(heap *h, void *d, void *aux)
{
    void *tmp;

    h->data[h->used++] = d;
    _fix_up(h, aux);

    if (h->used == h->size) {
        h->size = ((h->size + 1) * 2) - 1;
        /* printf("DEBUG: resizing\n");  */
        tmp = xrealloc(h->data, h->size * sizeof(void *));
        if (tmp != NULL)
            h->data = tmp;
    }

    return 0;
}

void *
heap_get(heap *h, void *aux)
{
    void *retval;
    
    if (h->used > 0) {
        retval = h->data[0];
        h->data[0] = h->data[h->used - 1];
        --h->used;
        _fix_down(h, aux);
    } else {
        retval = NULL;
    }

    if (h->get_hook != NULL)
        h->get_hook(retval, h->get_hook_aux);
    return (retval);
}

void *
heap_peek(heap *h)
{
    void *retval;

    if (h->used > 0)
        retval = h->data[0];
    else
        retval = NULL;
    if (h->peek_hook != NULL)
        h->peek_hook(retval, h->peek_hook_aux);

    return (retval);
}

void
heap_iterate(heap *h, void (*func)(void *, void *), void *aux)
{
    _heap_iterate_gen(h, func, aux);
}

void
heap_destroy(heap *h, void (*func)(void *, void *), void *aux)
{
    _heap_iterate_gen(h, func, aux);
    free(h->data);
    free(h);
}


void
_heap_iterate_gen(heap *h, void (*func)(void *, void *), void *aux)
{
    int i;
    
    for (i = 0; i < h->used; ++i) 
        func(h->data[i], aux);
}

void 
heap_set_put_hook(heap *h, void (*func)(void *, void *), void *aux)
{
    h->put_hook = func;
    h->put_hook_aux = aux;
}

void
heap_set_get_hook(heap *h, void (*func)(void *, void *), void *aux)
{
    h->get_hook = func;
    h->get_hook_aux = aux;
}

void
heap_set_peek_hook(heap *h, void (*func)(void *, void *), void *aux)
{
    h->peek_hook = func;
    h->peek_hook_aux = aux;
}


void
heap_debug_dump(heap *h, void (*func)(void *))
{
    char *indent;

    printf("Debug dump:\n");
    func(h->data[0]); 
    puts("");

    indent = (char *) alloca(strlen(" ") + 1);
    strcpy(indent, " ");
    _heap_debug_dump_recursive(h, func, 1, indent);
    _heap_debug_dump_recursive(h, func, 2, indent);
    printf("Debug dump done.\n");

}

void
_heap_debug_dump_recursive(heap *h, void (*func)(void *), int level, char *s)
{
    char *indent;
    if (level < h->used) {
        printf("%s", s);
        func(h->data[level]);
        puts("");
        indent = (char *) alloca(2 * strlen(s) + 1);
        strcpy(indent, s); 
        strcat(indent, s);
        _heap_debug_dump_recursive(h, func, level * 2 + 1, indent);
        _heap_debug_dump_recursive(h, func, level * 2 + 2, indent);
    }
}


/* The functions fix_up() and fix_down() have the job of maintaining the heap
   invariant.  fix_up is a transliteration from Sedgewick to readable C.

   fix_up is presented in Sedgewick as fixUp [sic].  */

#define SWAP(n, m) {void *__tmp; __tmp = n; n = m; m = __tmp;}

void 
_fix_up(heap *h, void *aux)
{
    int size;

    for (size = h->used; size > 1; size /= 2) {
        if (h->cmp(h->data[size / 2 - 1], h->data[size - 1], aux) < 0)
            SWAP(h->data[size - 1], h->data[size / 2 - 1]);
    }
}

void
_fix_down(heap *h, void *aux)
{   
    if (h->used >= 2 && h->cmp(h->data[0], h->data[1], aux) < 0) {
        SWAP(h->data[0], h->data[1]);
        _fix_down_recursive(h, aux, 1);

    }
    if (h->used >= 3 && h->cmp(h->data[0], h->data[2], aux) < 0) {
        SWAP(h->data[0], h->data[2]);
        _fix_down_recursive(h, aux, 2);
    }
}   
    
void
_fix_down_recursive(heap *h, void *aux, int level)
{   
    if (level * 2 + 1 <= h->used) {
        if (h->cmp(h->data[level], h->data[level * 2 + 1], aux) < 0) {
            SWAP(h->data[level], h->data[level * 2 + 1]);
            _fix_down_recursive(h, aux, level * 2 + 1);
        }
    }
    if (level * 2 + 2 <= h->used) {
        if (h->cmp(h->data[level], h->data[level * 2 + 2], aux) < 0) {
            SWAP(h->data[level], h->data[level * 2 + 2]);
            _fix_down_recursive(h, aux, level * 2 + 2);
        }
    }
}   
