/*
 * Simple, robust hash table implementation.  
 *
 * Copyright (c) 2001 Michael Wolf (mike@dodgysoftware.net)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* $Header: /cvsroot/moodns/moodns/src/ht.c,v 1.1.1.1 2001/12/21 02:43:02 mawolf Exp $  */

#include <stdlib.h>
#include <string.h>

#include "ht.h"

/* Yuk!  */
#if (!defined(BJENKINS_HASH_FUNC)) && (!defined(PRACTICE_PROG_HASH_FUNC))
# define PRACTICE_PROG_HASH_FUNC
#endif



static const unsigned int SIZE = 101;
static const float LOAD = (float) 3.0 / 4.0;

static unsigned int hash(unsigned char *, unsigned long int);
static void _ht_iterate_gen(ht *, void (*)(char *, void *, void *), void *, 
                            int);
static void _ht_rehash(ht *);

struct list {
    char *key;
    void *data;
    struct list *next;
};

struct ht {
    /* The number of buckets in the table.  */
    int size;
    /* The number of non-empty buckets in the table.  */
    int filled;
    /* The table itself.  */
    struct list **data;
    /* Comparison function.  */
    int (*cmp)(void *a, void *b, void *c);
    /* A function to delete the old element after a new one's been added.  */
    void (*del)(void *a, void *b);
    /* A function that's automagically invoked when ht_get() is called.  */
    void (*get_hook)(char *key, void *data, void *aux);
    /* Auxiliary data for get_hook.  */
    void *get_hook_aux;
    /* A function that's automagically invoked when ht_put() is called.  */
    void (*put_hook)(char *key, void *data, void *aux);
    /* Auxiliary data for put_hook.  */
    void *put_hook_aux;
};

ht *
ht_init(int (*compare)(void *, void *, void *), void (*delete)(void *, void *))
{
    ht *retval;

    retval = (ht *) calloc(1, sizeof(ht));
    retval->size = SIZE;
    retval->filled = 0;
    retval->data = (struct list **) calloc(retval->size, sizeof(struct list *));
    retval->cmp = compare;
    retval->del = delete;
    retval->get_hook = NULL;
    retval->get_hook_aux = NULL;
    retval->put_hook = NULL;
    retval->put_hook_aux = NULL;
    return (retval);
}

int
ht_put(ht *table, char *key, void *value, void *del_arg)
{
    void *retval;
    unsigned int hash_val;
    struct list *l, *n;
    int found;

    hash_val = hash(key, table->size);
    hash_val = hash_val % table->size;
    if (table->data[hash_val] == NULL) {
        table->data[hash_val] = (struct list *) calloc(1, sizeof(struct list));
        table->data[hash_val]->key = strdup(key);
        table->data[hash_val]->data = value;
    } else {
        /* Iterate the chained list until the key is found or the 
           list ends.  */
        found = 0;
        for (l = table->data[hash_val]; l != NULL; l = l->next) {
            if (!strcmp(l->key, key)) {
                found = 1;
                break;
            }
        }
        if (found == 0) {
            n = (struct list *) calloc(1, sizeof(struct list));
            n->key = strdup(key);
            n->data = value;
            n->next = table->data[hash_val];
            table->data[hash_val] = n;
            ++table->filled;
        } else {
            (table->del)(l->data, del_arg);
            retval = l->data;
            l->data = value;
        }
    }

    if ((float) table->filled / table->size > LOAD)
        _ht_rehash(table);
    if (table->put_hook != NULL) {
        (table->put_hook)(key, value, table->put_hook_aux);
    }
    return (0);
}

void *
ht_get(ht *table, char *key)
{
    void *retval;
    unsigned int hash_val;
    int found;
    struct list *l;
    
    hash_val = hash(key, table->size);
    hash_val = hash_val % table->size;
    if (table->data[hash_val] == NULL) {
        retval = NULL;
    } else {
        found = 0;
        for (l = table->data[hash_val]; l != NULL; l = l->next) {
            if (!strcmp(l->key, key)) {
                found = 1;
                break;
            }
        }
        if (found == 0) {
            retval = NULL;
        } else {
            retval = l->data;
        }
    }
    /*
     * If table->get_hook != NULL, then get_hook is run, even if retval == NULL
     */
    if (table->get_hook != NULL ) {
        (table->get_hook)(key, retval, table->get_hook_aux);
    }
    return (retval);
}

void *
ht_remove(ht *table, char *key)
{
    void *retval;
    unsigned int hash_val;
    int found;
    struct list *l, *m;

    retval = NULL;
    hash_val = hash(key, table->size);
    hash_val = hash_val % table->size;
    if (table->data[hash_val] == NULL) {
        /* Nothing to remove.  */
    } else {
        found = 0;
        for (l = table->data[hash_val]; l != NULL; l = l->next) {
            if (!strcmp(l->key, key)) {
                found = 1;
                break;
            }
        }
        if (found == 0) {
            /* Well, it's not there.  Oh well.  */
        } else {
            /* Remove the element from the list.  */
            if (l == table->data[hash_val]) {
                table->data[hash_val] = l->next; /* Could be NULL */
            } else {
                m = table->data[hash_val];
                while (m->next != l) 
                    m = m->next;
                m->next = l->next;
            }
            /* Now deallocate l.  */
            free(l->key);
            retval = l->data;
            free(l);
            --table->filled;
        }
    }
    return (retval);
}

void
ht_destroy(ht *table, void (*finalise)(char *, void *, void *), void *aux)
{
    _ht_iterate_gen(table, finalise, aux, 1);
}

void
ht_iterate(ht *table, void (*func)(char *, void *, void *), void *aux)
{
    _ht_iterate_gen(table, func, aux, 0);
}

void
ht_set_get_hook(ht *table, void (*func)(char *, void *, void *), void *get_aux)
{
    table->get_hook = func;
    table->put_hook_aux = get_aux;
}

void
ht_set_put_hook(ht *table, void (*func)(char *, void *, void *), void *put_aux)
{
    table->put_hook = func;
    table->put_hook_aux = put_aux;
}

void
_ht_iterate_gen(ht *table, void (*func)(char *, void *, void *), void *aux,
                 int dealloc)
{
    int i;
    struct list *l, *m;
    for (i = 0; i < table->size; ++i) {
        l = table->data[i];
        while (l != NULL) {
            m = l;
            l = l->next;
            func(m->key, m->data, aux);
            if (dealloc != 0)
                free(m);
        }
    }
    if (dealloc != 0) {
        free(table->data);
        free(table);
    }
}

void
_ht_rehash(ht *table)
{
    int newsize;
    struct list **newdata;
    struct list *old;
    struct list *tmp;
    int i;
    unsigned int hash_val;


    newsize = table->size * 2 + 1;
    newdata = (struct list **) calloc(newsize, sizeof(struct list *));
    if (newdata != NULL) {
        for (i = 0; i < table->size; ++i) {
            old = table->data[i];
            while (old != NULL) {
                tmp = old;
                old = old->next;
                hash_val = hash(tmp->key, newsize);
                hash_val %= newsize;
                tmp->next = NULL;
                if (newdata[hash_val] == NULL) {
                    ;
                } else {
                    tmp->next = newdata[hash_val];
                }
                newdata[hash_val] = tmp;
            }
        }
        free(table->data);
        table->size = newsize;
        table->data = newdata;
    }
}


/* Functions which may be of use when debugging ht or programs which use
   ht.  */
#include <stdio.h>
void
ht_debug_dump(ht *table, void (*print_func)(void *))
{
    struct list *placeholder;
    int i;

    for (i = 0; i < table->size; ++i) {
        printf("Bucket %d:\n", i);
        for (placeholder = table->data[i]; placeholder != NULL; 
             placeholder = placeholder->next) {
            if (print_func == NULL) {
                printf("\t%s => %p\n", placeholder->key, placeholder->data);
            } else {
                printf("\t%s => ", placeholder->key);
                (*print_func)(placeholder->data);
                printf("\n");
            }
        }
        printf("\n");
    }
}



/* I'm less sure about the effectiveness of this hash function than I was.
   Output from ht_debug_dump() suggests one of three possibilities (that I can 
   think of)
   - The hash function isn't so good
   - I'm using it incorrectly somehow
   - I've been having horrible luck with keys that happen to be similar and
     happen to hash the same.  
   I think the best thing to do now is to try out some different hashing
   functions, see how they work, maybe run some benchmarks.  Admittedly, the 
   real use of a key=>value type structure is the ease of programming it 
   provides - using ht is really easy and has been a total win so far - but
   the poor behaviour characteristics have been a bit disturbing.  

   In test2.c, where I try using somewhat more random keys, the distribution
   seems to be a lot better.  It would probably help if I understood what
   the hashing function actually does.  

   Some other hashes to try: 
   - Others from Bob Jenkins' site
   - The one in Practice of Programming
   - Something from Knuth
   - Something else from the net, but there isn't much

   Hmmm.  */

/* Sun 12 Aug 15:19: I copied in the hash function from _The Practice of 
   Programming_ and set that to the default.  Empirically, it seems to give
   a better distribution than the function I had previously been using.  

   It would probably be a good idea to try out some other hashes too, but I
   don't see that as critical at this point.  */

#ifdef BJENKINS_HASH_FUNC

/* I found the hash function below at
       http://burtleburtle.net/bob/hash/doobs.html
   I've massaged it a bit to fit with my coding style.  The URL explains
   what's going on better than I can.  The mix() macro is from the same place,
   and is complementary to the hash() function.  */
#define mix(a, b, c) \
{ \
  a -= b; a -= c; a ^= (c >> 13); \
  b -= c; b -= a; b ^= (a << 8); \
  c -= a; c -= b; c ^= (b >> 13); \
  a -= b; a -= c; a ^= (c >> 12); \
  b -= c; b -= a; b ^= (a << 16); \
  c -= a; c -= b; c ^= (b >> 5); \
  a -= b; a -= c; a ^= (c >> 3); \
  b -= c; b -= a; b ^= (a << 10); \
  c -= a; c -= b; c ^= (b >> 15); \
}



unsigned int
hash(unsigned char *k, unsigned long int initval)
{
    register unsigned long int a, b, c, len;

    /* Set up the internal state */
    len = strlen(k) - 1;
    a = b = 0x9e3779b9;     /* the golden ratio; an arbitrary value */
    c = initval;        /* the previous hash value */

    /* handle most of the key */
    while (len >= 12) {
        a += (k[0] + ((unsigned long int) k[1] << 8) + ((unsigned long int) k[2]
 << 16) + ((unsigned long int) k[3] << 24));
        b += (k[4] + ((unsigned long int) k[5] << 8) + ((unsigned long int) k[6]
 << 16) + ((unsigned long int) k[7] << 24));
        c += (k[8] + ((unsigned long int) k[9] << 8) + ((unsigned long int) k[10
] << 16) + ((unsigned long int) k[11] << 24));
        mix(a, b, c);
        k += 12;
        len -= 12;
    }

    /* handle the last 11 bytes */
    c += len;
    switch (len) {      /* all the case statements fall through */
    case 11:
        c += ((unsigned long int) k[10] << 24);
    case 10:
        c += ((unsigned long int) k[9] << 16);
    case 9:
        c += ((unsigned long int) k[8] << 8);
        /* the first byte of c is reserved for the length */
    case 8:
        b += ((unsigned long int) k[7] << 24);
    case 7:
        b += ((unsigned long int) k[6] << 16);
    case 6:
        b += ((unsigned long int) k[5] << 8);
    case 5:
        b += k[4];
    case 4:
        a += ((unsigned long int) k[3] << 24);
    case 3:
        a += ((unsigned long int) k[2] << 16);
    case 2:
        a += ((unsigned long int) k[1] << 8);
    case 1:
        a += k[0];
        /* case 0: nothing left to add */
    }
    mix(a, b, c);
    /* report the result */
    return (c);
}


#endif

#ifdef PRACTICE_PROG_HASH_FUNC


/* This is the hash function given in Kernighan and Pike's _The Practice of 
   Programming_, pp 56-57.  It has an extra parameter.  */


enum _hash2_blah {
    MULTIPLIER = 31
};

unsigned int
hash(unsigned char *str, unsigned long int size)
{
    unsigned int h;
    unsigned char *p;

    h = 0;
    for (p = (unsigned char *) str; *p != '\0'; p++) 
        h = MULTIPLIER * h + *p;

    return (h % size);
}


#endif

