/* Copyright (c) 2001, 2002 Omar Kilani, Michael Wolf.  All rights reserved.  
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * moodns is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Header: /cvsroot/moodns/moodns/src/lookup.c,v 1.3 2002/06/19 13:34:12 mawolf Exp $  */

/* This file contains code which does a few easy things and one fairly hard
   thing.  

   There is code to look up records from the database and from cache.  Those
   are easy.

   There is also code to look up records which aren't in either the cache or
   (our) database.  This is hard.  Here's how it works.  (For simplicity of
   explanation, we assume that there are no redundant servers.)

   We want to find out the IP address (A record) of dodgysoftware.net.  

   - Ask the cache for the A record of dodgysoftware.net, which fails.

   - Ask the cache for the NS record of dodgysoftware.net, which fails.

   - Ask the cache for the NS record of net, which fails.  

   - Ask the root server for the NS record of net.  Hypothetical answer:
   ns1.net.  Ask the root server what the A record of ns1.net is.  
   Hypothetical answer: 192.168.0.2
   
   - Ask 192.168.0.2 for the NS record(s) of dodgysoftware.net.  Hypothetical
   answer: ns1.dodgysoftware.net.  Ask 192.168.0.2 for the A record of
   ns1.dodgysoftware.net.  Hypothetical answer: 192.168.5.1.  

   - Ask 192.168.5.1 for the A record of dodgysoftware.net.  Hypothetical
   answer: 192.168.34.35.  

   - We're done.  Return the answer to our client.  Inform the cache of what
   we learned.  

   NB: There may be differences, however, possibly due to a misunderstanding
   of the protocol on my part or due to different configurations.  

   - Along with a NS record, we might also get an Additional section with 
   info we could use, like A records.  

   - Of course, the cache may have some of the intermediate information.  In 
   that case, we break out of the recursion early.  


   This is a more general description of the algorithm:  

   We're looking for the X record of domain name D.  
   We'll call the parent domain of D D' (D prime).  When we set D to D', D' 
   is implicitly set to D''.  
   We'll call the child domain we're interested in, if any, C.  
   We'll use functional notation for record types; eg the NS record of D can
   be denoted NS(D).  
   Assume for the moment that none of the information we want it cached.  
   Also assume that there are no redundant servers (if there were, we'd 
   pick one at random).  

   1. Check for a NS record for D.  If not found and D' is non-null [*], 
   set D to D', C to D.  

   2. (Now we're talking to a root server.)  Obtain NS(C).  Obtain A(NS(C))
   (This may be included in the Additional section of a packet, or we may
   need to explicitly ask for it.)  
   
   3. 


   10. 



   [*] If D is foo.com, D' is com, non-null.  If D is com, D' is null.


   Arg.  Hard to explain.  I'll try to code it first.  Maybe I'll be able to
   code it without being able to describe it in words.  Arg.



   

   (This is probably described in doc/rfc1034.txt or doc/rfc1035.txt, but I 
   haven't actually found that description.)
*/ 

/* Known issues:

   - Indirect lookup doesn't work.  But I'm close.  

   - There is HEAPS of redundant code that needs to be factored out.  And HEAPS
   of code that's just plain messy and needs to be cleaned up.  

*/

#ifndef _REENTRANT
#define _REENTRANT
#endif

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "misc.h"
#include "protocol.h"
#include "database.h"
#include "options.h"
#include "cache.h"
// #include "wpthread.h"
// #include "byte_array.h"
// #include "socket.h"
#include "query.h"
#include "lookup.h"

#include "xmalloc.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif

extern struct global_options globals;

extern cache *c;
extern database *d;
extern record_chain *roots;

typedef struct minilist minilist;
struct minilist {
    char *name;
    minilist *next;
};

static record_chain *recursive_lookup(char *, enum record_type);
static record_chain *real_recursive_lookup(char *, enum record_type, minilist *);
static record_chain *lookup_from_root_servers(minilist *);
void load_an(packet *, record_chain *, record_chain *);
void load_ns(packet *, record_chain *, record_chain *);
void load_ar(packet *, record_chain *, record_chain *);

static int is_parent(char *, char *);

int
indirect_lookup(packet *pa)
{
    char *rn;
    int i, n;
    struct record_chain *rc;
    struct resource_record *rr;
    
    for (i = 0; i < pa->h->qdcount; ++i) {
        rn = pa->qd[i]->qname;
        rc = cache_check(c, rn, pa->qd[i]->qtype);
        if (rc != NULL) {
            for (n = 0; n < rc->records_len; ++n) {
                rr = record_to_resource_record(&rc->records[n]);
                pa->h->ancount++;
                pa->an = xrealloc(pa->an, pa->h->ancount *
                                  sizeof(struct resource_record *));
                pa->an[pa->h->ancount - 1] = rr;
                pa->size_in_bytes += resource_record_size(rr);
            }
        } else {
            /* We need to look it up the hard way.  */            
            rc = recursive_lookup(rn, pa->qd[i]->qtype);
        }
    }
    
    return 0;
}

/* Set the stage for real_recursive_lookup.  */

struct record_chain *
recursive_lookup(char *name, enum record_type type)
{
    minilist *a, *b;
    char *x;
    char *n = name;
    a = (minilist *) alloca(sizeof(minilist));
    a->next = NULL;
    b = a;
    
    b->name = strdupa(n); printf("copying %s\n", name);
    while ((x = strchr(n, '.')) != NULL) {
        b->next = (minilist *) alloca(sizeof(minilist));
        b->next->next = NULL;
        b = b->next;
        n = x + 1;
        b->name = strdupa(n); printf("copying %s\n", n);
    }
    
    return real_recursive_lookup(name, type, a);
}

/* This function returns a record chain containing NS(dom->name) and 
   A(NS(dom->name)).  */
   
record_chain *
real_recursive_lookup(char *name __attribute__((unused)),
                      enum record_type type __attribute__((unused)), 
                      minilist *dom __attribute__((unused)))
{
    record_chain *xns, *a, *aa;
    record_chain *necessary;
    int i;
    packet *p, *p1;
    u16 id;
    
    if (dom->next == NULL) {
        return lookup_from_root_servers(dom);
    } else {
        xns = cache_check(c, dom->name, NS);
        if (xns == NULL) {
            necessary = real_recursive_lookup(name, type, dom->next);
            if (necessary == NULL)
                goto cleanup;
        } else {
            a = record_chain_init(dom->name);
            for (i = 0; i < xns->records_len; ++i) {
                aa = cache_check(c, xns->records[i].record.ns, A);
                record_chain_add_chain(a, aa);
            }
            if (a->records_len == 0) {
                /* Well, we're out of luck.  */
                record_chain_destroy(a);
                return NULL;
            }
            necessary = a;
        }
        p = (packet *) xcalloc(1, sizeof(packet));
        p->h = (struct header *) xcalloc(1, sizeof(struct header));
        packet_add_question(p, dom->name, NS, IN);
        for (i = 0; i < necessary->records_len; ++i) {
            record_chain *an;
            record_chain *ns;
            record_chain *ar;
            record *rec;
            int x;
            id = (u16) (random() & 0xFF);
            p->h->id = id;
            p1 = query_send(p, necessary->records[i].record.a, 53, 10);
            if (!p1->h->ancount && !p1->h->nscount && !p1->h->arcount)
                goto cleanup;
            /* if (p1->h->id != id)
               goto cleanup;  */
            an = record_chain_init(dom->name);
            for (x = 0; x < p1->h->ancount; ++x) {
                rec = resource_record_to_record(p1->an[x]);
                record_chain_add(an, rec);
                record_delete(rec);
            }
            ns = record_chain_init(dom->name);
            for (x = 0; x < p1->h->nscount; ++x) {
                rec = resource_record_to_record(p1->ns[x]);
                record_chain_add(ns, rec);
                record_delete(rec);
            }
            ar = record_chain_init(dom->name);
            for (x = 0; x < p1->h->arcount; ++x) {
                rec = resource_record_to_record(p1->ar[x]);
                record_chain_add(ar, rec);
                record_delete(rec);
            }
            
            if (ar->records_len > 0) {
                /* assume that the records in ar correspond to the records
                   in an  */
                a = ar;
            }

            packet_free(p);
            packet_free(p1);
            break;
        }
        // rchoice = random() & a->records_len;
        
        return a;
    }
cleanup:
    if (p)
        packet_free(p);
    if (p1)
        packet_free(p1);
    return NULL;
}

/* I don't know if a server is required to give the A records you're 
   "likely" to want based on corresponding NS requests.  In fact, I
   don't know if the server is required to have those records at all!  
   (Although obviously it should.)  I don't know if the A records it gives
   will necessarily correspond to the NS records it's given.  I don't even
   know if the server will put the right kinds of responses in the right
   parts of the packet!  So we need to be prepared against all sorts of 
   brokenness.  */
record_chain *
lookup_from_root_servers(minilist *dom)
{
    record_chain *retval;
    packet *p, *result;
    int rchoice;
    u32 remote;
    u16 id;
    record_chain *ns, *a;
        
    assert(roots->records_len > 0);

    // retval = NULL;

    id = (u16) (random() & 0xFF);
    p = (packet *) xcalloc(1, sizeof(packet));
    p->h = (struct header *) xcalloc(1, sizeof(struct header));
    p->h->id = id;
    p->h->opcode = 0;
    
    packet_add_question(p, dom->name, NS, IN);
    rchoice = random() & (roots->records_len - 1);
    remote = roots->records[rchoice].record.a;

    result = query_send(p, remote, 53, 10);
    if (!result) 
        goto done;
    /* if (result->h->id != id) 
       goto done;  */
    if (!result->h->ancount && !result->h->nscount && !result->h->arcount) 
        goto done;
    ns = record_chain_init(xstrdup(dom->name));
    a = record_chain_init(xstrdup(dom->name));
    
    load_an(result, ns, a);
    load_ns(result, ns, a);
    load_ar(result, ns, a);
    
    if (a->records_len == 0) {
        /* Ask the root server for some A records.  Hm.  */
    }

    /* XXX: Now check record chains for consistency as described above.  */
    
    cache_put_record_chain(c, ns);
    cache_put_record_chain(c, a);
    retval = a;
done:
    /* FIXME: deallocate the following: p, result, ns  */
    return retval;
}

void
load_an(packet *p, record_chain *ns, record_chain *a)
{
    int i;
    record *rec;

    assert(p != NULL); 
    assert(ns != NULL);
    assert(a != NULL);

    for (i = 0; i < p->h->ancount; ++i) {
        rec = resource_record_to_record(p->an[i]);
        switch (rec->type) {
        case NS:
            record_chain_add(ns, rec);
            break;
        case A:
            record_chain_add(a, rec);
            break;
        default:
            /* XXX: What do do?  Ignore it?  */
            break;
        }
        record_delete(rec);
    }
}

void
load_ns(packet *p, record_chain *ns, record_chain *a)
{
    int i;
    record *rec;

    assert(p != NULL); 
    assert(ns != NULL);
    assert(a != NULL);

    for (i = 0; i < p->h->nscount; ++i) {
        rec = resource_record_to_record(p->ns[i]);
        switch (rec->type) {
        case NS:
            record_chain_add(ns, rec);
            break;
        case A:
            record_chain_add(a, rec);
            break;
        default:
            break;
        }
        record_delete(rec);
    }
}


void
load_ar(packet *p, record_chain *ns, record_chain *a)
{
    int i;
    record *rec;

    assert(p != NULL); 
    assert(ns != NULL);
    assert(a != NULL);

    for (i = 0; i < p->h->arcount; ++i) {
        rec = resource_record_to_record(p->ar[i]);
        switch (rec->type) {
        case NS:
            record_chain_add(ns, rec);
            break;
        case A:
            record_chain_add(a, rec);
            break;
        default:
            break;
        }
        record_delete(rec);
    }
}

int 
direct_lookup(packet *pa)
{
    char *rn;
    int i, n;
    struct record_chain *rc;
    struct resource_record *rr;
    
    for (i = 0; i < pa->h->qdcount; ++i) {
        rn = pa->qd[i]->qname;
        rc = d->get_records(d, rn);
        if (rc != NULL) {
            for (n = 0; n < rc->records_len; ++n) {
                if (rc->records[n].type == pa->qd[i]->qtype) {
                    rr = record_to_resource_record(&rc->records[n]);
                    pa->h->ancount++;
                    pa->an = xrealloc(pa->an, pa->h->ancount * 
                                      sizeof(struct resource_record *));
                    pa->an[pa->h->ancount - 1] = rr;
                    pa->size_in_bytes += resource_record_size(rr);
                }
            }
        }
    }
    
    return 0;
}


/* Is potential parent pp a parent domain of potential child pc?  */

int
is_parent(char *pp, char *pc)
{
    /* Heh heh heh.  There's always a more better way to shut gcc up.  */
    return is_parent(pp, pc);
}






#ifdef TEST

cache *c;
database *d;

/**************************************************************************
 *                                                                        *
 *  This is a main routine.  (Do you miss sickeningly obvious statements  *
 *  in comments in pedagogical programs?)                                 *
 *                                                                        *
 **************************************************************************/
int
main(int argc, char **argv)
{
    char *domain = "osgiliath.schemes.dodgysoftware.net";
    
    bottom_up(domain);


    return 0;
}

#endif
