/*
 * The merrimack process id file library.
 *
 * Copyright (c) 2001 Michael Wolf (mike@dodgysoftware.net)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
/*
 * For a more thorough treatment, see 
 *     http://dodgysoftware.net/ironclad/merrimack.html
 *
 * Background: many long running programs, such as servers, only want to be 
 * running in one instance.  This is generally because they rely on exclusive
 * access to a resource, such as a TCP port.  A common technique to realise
 * this is to create a ``pid-file'' storing the process ID of the currently
 * running program.  When a new instance of that program is started, it will
 * look for that pid-file and assert that the process id contained there
 * is not running.
 *
 * While many programs do indeed do this, most of them do not do it properly.
 * This library, merrimack, is an attempt to make proper pid-files easy to 
 * use and manipulate.  At the same time, it attempts to be portable; the same
 * interface to merrimack is intended to work on any Unix or Unix-like system.
 */


/* $Header: /cvsroot/moodns/moodns/src/merrimack.c,v 1.2 2002/07/20 16:57:06 mawolf Exp $  */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <syslog.h>

#include "util.h"
#include "merrimack.h"

#define ERROR_SET(n, m) if (n != NULL) *n = m;

static char *create_filename(char *, pid_t);

struct mpid {
    char *program_name;
    char *pid_path;
    int fd;
};

int
mpid_create(char *name, pid_t pid, int *errors)
{
    /* The proper way to create a lockfile seems to be:  
       1) Create a file name.host.pid.
       2) Link that to name.
       3) Check that the link count on name == 2 - if not, unlink name, 
       unlink name.host.pid, and signal failure.
       4) Done.

       If any of the above fails, be sure to cleanup.
    */
    int retval;
    char *filename;
    int ret;
    FILE *x;
    struct stat sb;
    mode_t foo;
    // foo = umask(000);
    foo = umask(022);
    /* foo = umask(777);  */
    
    filename = create_filename(name, pid);
    if (filename == NULL) {
        ERROR_SET(errors, WEIRDNESS);
        retval = 0;
        goto done;
    }

    // ret = open(filename, O_CREAT | O_EXCL);
    x = fopen(filename, "w");    
    if (x == NULL) {
        ERROR_SET(errors, PERMISSION);
        retval = 0;
        goto done;
    }
    
    ret = link(filename, name);
    if (ret == -1) {
        ERROR_SET(errors, WEIRDNESS); /* XXX: This is an oversimplification  */
        printf_on_debug("link failed: %s\n", strerror(errno));
        retval = 0;
        goto done;
    }
    ret = stat(name, &sb);
    if (sb.st_nlink != 2) {
        ERROR_SET(errors, ALREADY);
        retval = 0;
        ret = unlink(filename);
        if (ret == -1)
            abort();
        ret = unlink(name);
        printf("already running\n");
        goto done;
    }
    

    retval = 1;
done:
    return retval;
}

int
mpid_destroy(char *name, pid_t pid, int *errors)
{
    /* The proper way to delete a lockfile seems to be:
       1) Unlink that name.
       2) Unlink name.host.pid.
       3) Done.
    */
    int retval;
    char *filename;
    int ret;
    
    filename = create_filename(name, pid);
    if (filename == NULL) {
        retval = 0;
        goto done;
    }
    
    ret = unlink(name);
    if (ret == -1) {
        ERROR_SET(errors, WEIRDNESS);
        retval = 0;
        goto done;
    }
    ret = unlink(filename);
    if (ret == -1) {
        ERROR_SET(errors, WEIRDNESS);
        retval = 0;
        goto done;
    }

    retval = 1;
done:
    return retval;
}
    
// int x;

//     printf("in mpid_destroy()\n");
//     // assert(mp != NULL);
//     x = close(mp->fd);
//     if (x == -1) {
//         perror("close failed");        
//     }
//     free(mp->program_name);
//     unlink(mp->pid_path);
//     free(mp->pid_path);
//     mp->fd = -1;
//     free(mp); 
//     return 0;
// }

#if 0
int
mpid_status(char *name, char *path, int *p)
{
    int retval;
    FILE *f;
    pid_t pid;
    int rv;

    f = fopen(path, "r");
    if (f == NULL) {
        /* It's not running.  */
        retval = 1;
    } else {
        pid = 0;
        fscanf(f, "%d", &pid);
        if (pid == 0) {
            /* Something weird  */
            retval = 2;
        } else {
            if (crufty_same(name, pid)) {
                if ((rv = kill(pid, 0)) == 0)
                    retval = 0;
                else
                    retval = 2;
            } else {
                retval = 3;
            }
        }
    }
    if (p != NULL)
        *p = pid;
    return (retval);
}

#endif

char *
create_filename(char *name, pid_t pid) 
{
    char *retval;
    char hostname[1024];
    size_t hostname_len;
    int ret;

    hostname_len = 1024;  /* Plenty of space.  */
    memset(hostname, (int) '\0', sizeof(hostname));
    ret = gethostname(hostname, hostname_len);
    if (ret == -1) {
        retval = NULL;
        goto done;
    }
    retval = NULL;
    asprintf(&retval, "%s.%s.%d", name, hostname, pid);
            
done:
    return retval;
}


