/* Copyright (c) 2002 Michael Wolf.  All rights reserved.  */
/* Copyright (c) 2002 Ximian Inc.  All rights reserved.  */

/* $Header: /cvsroot/moodns/moodns/src/mooinet.c,v 1.2 2002/07/20 17:00:09 mawolf Exp $  */

/* usage:
   mooinet -p udp|tcp -o portnum -i numerical-ip -s path
   every option must be specified.  if this mooinet communicating with only
   an authoritative or a resolving server, set the other's path to /dev/null. 
   
   there must be one mooinet running per ip/port/protocol combination.  
   beside that, you can have as many or as few as you like and as make sense.
   use -i 0.0.0.0 to bind to all IPs available.
   
   if mooinet is running as root, it drops its root privileges after binding
   (or failing to bind!) the specified port.
*/

/* the protocol between mooinet and the backend servers:
   
in the case of normal messages, 
   
mooinet sends messages like:
u32:id, u32 ip, u32:length, data
   
mooinet receives messages like:
u32:id, u32 ip, u32:length, data
   
id for normal messages must be greater than 2.  
   
in the case of extra messages,
   
id 1 is reserved for messages from mooinet to a server and id 2 is 
reserved for messages from a server to mooinet.  id 0 is a message to 
reset the stream.  this part of the protocol needs work.  (maybe set 
timeout reads)
   
   
*/

/* long term, i would like to generalise this to be useful for purposes
   unrelated to moodns.  */

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <pwd.h>
#include <errno.h>
#include <assert.h>

#include "byte_array.h"
#include "socket.h"
#include "util.h"
#include "xmalloc.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif

#define MAX_DGRAM 512



enum connection_state {
    /* initial state */
    CONNECTED = 0, 
    /* waiting for the remote end to send something */
    WAITING_RECV_REMOTE,
    /* waiting to send to the local end */
    WAITING_SEND_LOCAL,
    /* waiting to receive data from the local end */
    WAITING_RECV_LOCAL,
    /* waiting to send to the local end */
    WAITING_SEND_REMOTE,
    /* finished */
    DONE
};

struct connection {
    enum connection_state state;
    socket_t *s;
    u8 buf[MAX_DGRAM];
};

#define CONNECTIONS_SIZE 10 /* for now */
struct connection_list {
    struct connection **connections;
    int connections_size;
    int connections_max; /* unused for now, maybe never */
};

enum fsm_retval {
    OK = 0,
    INCOMPLETE = 1,
    COMPLETE = 2,
    ERROR = 3
};

enum fsm_retval fsm(struct connection *, byte_array *);
struct connection *connection_init(socket_t *);

enum fsm_retval
fsm(struct connection *c, byte_array *d __attribute__((unused)))
{
    enum fsm_retval retval;

    switch (c->state) {
    case CONNECTED:
        printf_on_debug("fsm: fd %d connected, waiting on input\n", c->s->fd);
        c->state = WAITING_RECV_REMOTE;
        retval = OK;
        break;
    case WAITING_RECV_REMOTE: 
        printf_on_debug("fsm: fd %d waiting on remote end\n", c->s->fd);
        break; 
    case WAITING_SEND_LOCAL: 
        printf_on_debug("fsm: fd %d waiting to send to local end\n", c->s->fd);
        break;
    case WAITING_RECV_LOCAL: 
        printf_on_debug("fsm: fd %d waiting on from local end\n", c->s->fd);
        break; 
    case WAITING_SEND_REMOTE: 
        printf_on_debug("fsm: fd %d waiting to send to local end\n", c->s->fd);
        break;
    case DONE:
        printf_on_debug("fsm: fd %d connection completed\n\n", c->s->fd);
        socket_close(c->s, SOCKET_BOTH);
        break;
    default:
        croak("bogus state reached");
        break;
    }

    return retval;
}

struct connection *
connection_init(socket_t *s)
{
    struct connection *retval;
    retval = xcalloc(1, sizeof(struct connection));
    retval->state = CONNECTED;
    retval->s = s;
    return retval;    
}



struct socket_buffer {
    int used;
    socket_t *s;
    u8 buf[MAX_DGRAM];
    size_t buf_idx;
};

int protocol = -1;
u32 ip = INADDR_NONE;
u32 port = 0;
char *sockpath = NULL;

socket_t *incomings;
socket_t *outgoings;

socket_t *create_socket(void);
socket_t *incoming(void);
int loop(void);
static void alloc_new_socket_buffer(struct socket_buffer *, int *, int *, socket_t *);
/*static int find_socket_buffer(struct socket_buffer *, int *, socket_t *);*/
static void cleanup(void);
static void cleanup_and_exit(void);
static void cleanup_and_croak(void);
static void sighandler(int);

int
main(int argc, char **argv)
{
    int retval;
    int i;
    int c;
    struct in_addr ia;

    printf_on_debug("mooinet starting\n");

    signal(SIGINT, sighandler);
    signal(SIGTERM, sighandler);
    signal(SIGABRT, sighandler);
    signal(SIGSEGV, sighandler);
   
    while ((c = getopt(argc, argv, "p:o:i:s:")) != -1) {
        switch (c) {
        case 'p':
            if (!strcasecmp(optarg, "tcp"))
                protocol = 1;
            else if (!strcasecmp(optarg, "udp"))
                protocol = 2;
            else
                die("unknown protocol");
            break;
        case 'o':
            port = strtoul(optarg, NULL, 10); /* XXX: Check this value.  */
            break;
        case 'i':
            i = inet_aton(optarg, &ia);
            if (!i) 
                die("bogus ip address");
            ip = ia.s_addr;
            break;
        case 's':
            sockpath = optarg;
            break;
        default:
            die("unknown option");
            break;
        }
    }
    
    if (protocol == -1)
        die("protocol not set\n");
    if (ip == INADDR_NONE)
        die("ip address not set\n");
    if (port == 0)
        die("port not set\n");
    if (sockpath == NULL)
        die("socket path not set\n");
   
    outgoings = create_socket();
    
    if (protocol != -1)
        incomings = incoming();
    
    retval = loop();
    
    return retval;
}

socket_t *
create_socket()
{
    socket_address_t s;
    s.local.path = sockpath;
    return socket_listener(&s, SOCKET_LOCAL, SOCKET_STREAM);
}

socket_t *
incoming()
{
    socket_t *retval;
    socket_address_t s;
    memset(&s, 0, sizeof(s));
    s.internet.port = port;
    s.internet.ip = ip;
    if (protocol == 1) {
        retval = socket_listener(&s, SOCKET_INTERNET, SOCKET_STREAM);
    } else if (protocol == 2) {
        retval = socket_listener(&s, SOCKET_INTERNET, SOCKET_DATAGRAM);
    } else {
        croak("invalid socket type");
    }
    return retval;
}

#define MAX(a, b) (a) > (b) ? (a) : (b)
int
loop()
{
    fd_set r, w, e;
    struct timeval tv;
    int ret;
    struct socket_buffer *sb;
    int sb_size;
    int sb_idx;
    int i;
    
    assert(outgoings != NULL);
    assert(incomings != NULL);
    assert(outgoings->job == SOCKET_LISTENER);
    assert(incomings->job == SOCKET_LISTENER);

    sb_size = 10;
    sb_idx = 0;

    sb = (struct socket_buffer *) 
        xcalloc(sb_size, sizeof(struct socket_buffer));

    while (1) {
        FD_ZERO(&r);
        FD_ZERO(&w);
        FD_ZERO(&e);
        tv.tv_sec = 10;
        tv.tv_usec = 0;
        
        FD_SET(outgoings->fd, &r);
        FD_SET(incomings->fd, &r);
            /* Need to add the descriptors in sb too  */
        for (i = 0; i < sb_size; ++i) {
            if (sb[i].used == 1) 
                FD_SET(sb[i].s->fd, &r);
        }
        errno = 0;
            /* XXX: arg.  This max stuff is a pain in the ass.  Maybe it would
               be easier to use poll.  */
        ret = select(MAX(outgoings->fd, incomings->fd) + 1, &r, &w, &e, &tv);
        if (ret == -1) {
            if (errno == EBADF || errno == EINVAL) {
                printf("select failed: %s\n", strerror(errno));
                abort();
            }
            if (errno == EINTR) {
                continue;
            }
        } else if (ret == 0)
            continue;

        printf_on_debug("select got something\n");

        if (FD_ISSET(outgoings->fd, &r))
            alloc_new_socket_buffer(sb, &sb_size, &sb_idx, outgoings);
        for (i = 0; i < sb_size; ++i) {
            if (sb[i].used == 1 && FD_ISSET(sb[i].s->fd, &r) != 0) {
                size_t bl1, bl2;
                byte_array msg;
                memset(&msg, 0, sizeof(msg));
                byte_array_catu32(&msg, sb[i].s->fd);
                byte_array_catu32(&msg, sb[i].s->socket_address.internet.sin_addr.s_addr);
                bl1 = socket_recv_poll(sb[i].s, &msg, MAX_DGRAM);
                if (bl1 > 0)
                    bl2 = socket_send(outgoings, &msg);
                xfree(msg.data);
            }
        }
        if (FD_ISSET(outgoings->fd, &r)) {
            size_t bl1, bl2, bl3;
            byte_array msgi, msgo;
            memset(&msgi, 0, sizeof(msgi));
            memset(&msgo, 0, sizeof(msgo));
                /* XXX: need better error checking here  */
            bl1 = socket_recv_poll(outgoings, &msgi, 3 * sizeof(u32));
            if (bl1 == 0)
                continue;
            bl2 = socket_recv_poll(outgoings, &msgi, ((u32 *) msgi.data)[2]);
            byte_array_cats(&msgo, &msgi.data[3], msgi.len - 3 * sizeof(u32));
            bl3 = socket_send(incomings, &msgo);
            xfree(msgi.data);
            xfree(msgo.data);
        }
    }
    return 0;
}

void
alloc_new_socket_buffer(struct socket_buffer *ss, int *size, int *idx, socket_t *sock)
{
    int i;

    for (i = 0; i < *size; ++i) {
        if (ss[i].used == 0)
            break;
    }
    if (i == *size) {
        ss = (struct socket_buffer *) 
            xrealloc(ss, (*size * 2) * sizeof(struct socket_buffer));
        *idx = *size;
        *size *= 2;
    } else {
        *idx = i;
        ss[*idx].used = 1;
        ss[*idx].s = socket_accept(sock);
    }
        
}

void
cleanup()
{
    int ret;
    ret = socket_close(incomings, SOCKET_BOTH);
    ret = socket_close(outgoings, SOCKET_BOTH);
    ret = unlink(outgoings->socket_address.local.sun_path);
}

void
cleanup_and_exit()
{
    cleanup();
    exit(1);
}

void
cleanup_and_croak()
{
    cleanup();
    croak("");
}

void 
sighandler(int sig)
{
    switch (sig) {
    case SIGTERM:
        printf_on_debug("caught SIGTERM\n");
        cleanup_and_exit();
        break;
    case SIGINT:
        printf_on_debug("caught SIGINT\n");
        cleanup_and_exit();
        break;
    case SIGABRT:
        printf_on_debug("caught SIGABRT\n");
        signal(SIGABRT, SIG_DFL); /* uurk  */
        cleanup_and_croak();
        break;
    default:
        printf_on_debug("caught unknown signal!\n");
        printf_on_debug("attempting to cleanup!\n");
        signal(SIGABRT, SIG_DFL);
        cleanup_and_croak();
        croak("unknown signal handled?\n");
        break;
    }
   
       
}


/*int
  find_socket_buffer(struct socket_buffer *ss, int *size, socket_t *sock)
  {
  int i;

  for (i = 0; i < *size; ++i) {
  if (sock->fd == ss->s->fd)
  goto done;
        
  }
  i = -1;
  done:
  return i;
  }*/
