/* Copyright (c) 2001 Michael Wolf.  All rights reserved. 
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * moodns is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */




/* $Header: /cvsroot/moodns/moodns/src/options.c,v 1.10 2002/02/08 03:36:13 mawolf Exp $  */

/* Because this file is done before anything else like chroot()ing or dropping
   root privs, it's very important that the code here be carefully
   written and correct.  */

/* To add a configuration option, do two things.  First, create an entry in
   the definition of struct global_options in options.h.  Second, 
   register that entry in conf_init.  Feel free to choose a different type to
   store different configuration options.  If you do that, you'll need to
   create a new _set_conf_* function here, and properly prototype it.  */

#ifndef _REENTRANT
#define _REENTRANT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <ctype.h>
#include <pthread.h>

#include "config.h"

#include "ht.h"
#include "wpthread.h"
#include "options.h"

#include "xmalloc.h"
#if WITH_DMALLOC
# include <dmalloc.h>
#endif
#define CONF_DIR "conf"

extern void chomp(char *);

extern struct global_options globals;

/* Forward declaration.  Yeeuch.  */
struct wrapper;

static void conf_init(void);
static int read_options(const char *);
static int _cxdzifklat(struct wrapper *, const char *, const char *);
static int _set_conf_str(void *, const char *);
static int _set_conf_number(void *, const char *);  
static int _set_conf_ip(void *, const char *);
static int _set_conf_bool(void *, const char *);

static int _local_cmp(void *, void *, void *);
static void _local_del(void *, void *);

static struct wrapper *wrapper_init(int (*)(void *, const char *), void *);
static void wrapper_nuke(struct wrapper *);

/* static */ struct wrapper {
    int (*func)(void *, const char *);
    void *where;
};

static pthread_once_t once_control = PTHREAD_ONCE_INIT;
static ht *cont;



    
void
options_init()
{
    wpthread_once(&once_control, conf_init);
}
    
/* If C had a real macro system, this wouldn't need to be so goddamn ugly.  
   And I want lambdas, too, while we're at it.  */
#define CONF_ENTRY_NUMBER(where, name) \
    wrap = wrapper_init(_set_conf_number, &where); \
    ht_put(cont, xstrdup(name), wrap, NULL); 

#define CONF_ENTRY_STRING(where, name) \
    wrap = wrapper_init(_set_conf_string, &where); \
    ht_put(cont, xstrdup(name), wrap, NULL); 

#define CONF_ENTRY_BOOL(where, name) \
    wrap = wrapper_init(_set_conf_bool, &where); \
    ht_put(cont, xstrdup(name), wrap, NULL); 

void
conf_init()
{
    struct wrapper *wrap;
    cont = ht_init(_local_cmp, _local_del);

    puts("hi from conf_init()");
    memset(&globals, 0, sizeof(globals));
    
    CONF_ENTRY_NUMBER(globals.general.num_threads, "num_threads");
    
    wrap = wrapper_init(_set_conf_number, &globals.net.port);
    ht_put(cont, xstrdup("port"), wrap, NULL);
    
    wrap = wrapper_init(_set_conf_ip, &globals.net.ip);
    ht_put(cont, xstrdup("ip"), wrap, NULL);
    
    wrap = wrapper_init(_set_conf_bool, &globals.net.use_tcp);
    ht_put(cont, xstrdup("use_tcp"), wrap, NULL);
    
    wrap = wrapper_init(_set_conf_bool, &globals.net.use_udp);
    ht_put(cont, xstrdup("use_udp"), wrap, NULL);
    
    wrap = wrapper_init(_set_conf_bool, &globals.cache.use_cache);
    ht_put(cont, xstrdup("use_cache"), wrap, NULL);
    
    wrap = wrapper_init(_set_conf_number, &globals.cache.num_cached);
    ht_put(cont, xstrdup("num_cached"), wrap, NULL);

    wrap = wrapper_init(_set_conf_bool, &globals.cache.cache_use_logging);
    ht_put(cont, xstrdup("cache_use_logging"), wrap, NULL);

    wrap = wrapper_init(_set_conf_number, &globals.cache.cache_logging_interval);
    ht_put(cont, xstrdup("cache_logging_interval"), wrap, NULL);

    wrap = wrapper_init(_set_conf_str, &globals.cache.cache_logging_path);
    ht_put(cont, xstrdup("cache_logging_path"), wrap, NULL);

    wrap = wrapper_init(_set_conf_str, &globals.security.user_name);
    ht_put(cont, xstrdup("user"), wrap, NULL);

    wrap = wrapper_init(_set_conf_str, &globals.security.directory);
    ht_put(cont, xstrdup("directory"), wrap, NULL);

    wrap = wrapper_init(_set_conf_str, &globals.authoritative.datafile);
    ht_put(cont, xstrdup("datafile"), wrap, NULL);

    wrap = wrapper_init(_set_conf_number, &globals.authoritative.default_ttl);
    ht_put(cont, xstrdup("default_ttl"), wrap, NULL);

    wrap = wrapper_init(_set_conf_bool, &globals.general.use_modules);
    ht_put(cont, xstrdup("use_modules"), wrap, NULL);

    wrap = wrapper_init(_set_conf_str, &globals.general.modules_path);
    ht_put(cont, xstrdup("modules_path"), wrap, NULL);

    /* Now we set up the "reasonable defaults".  Other code that checks
       configuration options doesn't need to check whether a specific option 
       has been explicitly set. 

       Should some of these "reasonable defaults" will cause moodns to fail in 
       operation, with the idea being that the administrator must do some 
       configuration?  */

    globals.general.num_threads = 5;
    globals.security.user_name = NULL;
    globals.security.directory = NULL;
    globals.net.port = 53;
    globals.net.ip = 0;
    globals.net.use_tcp = 0;
    globals.net.use_udp = 1;
    globals.cache.use_cache = 1;
    globals.cache.num_cached = 2000;
    globals.authoritative.use_authoritative = 0;
    globals.authoritative.datafile = NULL;
    globals.authoritative.default_ttl = 86400;  /* One day.  */
    
    read_options(CONF_DIR);
}
    
int
_set_conf_str(void *_where, const char *str)
{
    char **where = (char **) _where;
    char *newstr = xstrdup(str);
    chomp(newstr);
    *where = newstr;
    return 0;
}
    
int
_set_conf_ip(void *_ip __attribute__((unused)), const char *str __attribute__((unused)))
{
    printf("unimplemented feature\n");
    abort();
    return 0;
}
    
int
_set_conf_bool(void *_alteree, const char *str)
{
    char *c;
    int truth;
    int *alteree = (int *) _alteree;
    
    c = (char *) alloca(strlen(str) + 1);
    strcpy(c, str);
    chomp(c);
    if (!strcasecmp(c, "true")) 
        truth = 1;
    else
        truth = 0;

    *alteree = truth;
    
    return truth;
}
    
int
_set_conf_number(void *_nimbur, const char *str)
{
    int retval;
    int num;
    int *nimbur = (int *) _nimbur;

    /* FIXME: We should check this a bit better.  */
    num = strtol(str, NULL, 10);

    *nimbur = retval = num;

    return 0;
}
    
    
int
read_options(const char *dir)
{
    int retval;
    DIR *p;
    struct dirent *n, *o;
    struct wrapper *wrap;
    
    n = (struct dirent *) alloca(sizeof(struct dirent));
    o = (struct dirent *) alloca(sizeof(struct dirent));
    memset(n, 0, sizeof(*n));
    memset(o, 0, sizeof(*o));

    p = opendir(dir);
    if (p == NULL) {
        printf("No configuration directory found.  Saya freekin' nara!\n");
        abort();
    }

    do {
        readdir_r(p, n, &o);
        wrap = ht_get(cont, n->d_name);
        if (wrap != NULL) {
            _cxdzifklat((struct wrapper *) wrap, dir, n->d_name);
        } else {
            if (strcmp(".", n->d_name) != 0 && strcmp("..", n->d_name) != 0) {
                printf("unknown thingo\n");
            } else if (strcmp("CVS", n->d_name) == 0) {
                printf("Ignoring CVS directory");
            }
        }
    } while (o != NULL);

    return retval = 0;
}


/* If you don't know what this stands for, you're not a good programmer.  */
int
_cxdzifklat(struct wrapper *wrap, const char *prefix, const char *name)
{
    char *path;
    int ret1, ret2;
    size_t len;
    struct stat sb;
    size_t file_size;
    int fd;
    char *contents;

    len = strlen(prefix) + strlen("/") + strlen(name) + 1;
    path = alloca(len);  
    memset(path, '\0', len);
    strcat(path, prefix);
    strcat(path, "/");
    strcat(path, name);
    memset(&sb, 0, sizeof(sb));

    ret1 = stat(path, &sb);
    if (ret1 == -1) {
        perror("stat() failed");
        abort();
    }

    file_size = sb.st_size;
    fd = open(path, O_RDONLY);
    if (fd == -1) {
        perror("open() failed");
        abort();
    }
    contents = mmap(NULL, file_size, PROT_READ, MAP_PRIVATE, fd, 0);
    if (contents == MAP_FAILED) {
        perror("mmap() failed");
        abort();
    }

    (*wrap->func)(wrap->where, contents);

    ret2 = munmap(contents, file_size);
    if (ret2 != 0) {
        perror("munmap() failed");
        abort();
    }
    
    return 0;
}

int
_local_cmp(void *a, void *b, void *d __attribute__((unused)))
{ 
    if (((struct wrapper *) a)->func > ((struct wrapper *) b)->func)  
        return 1; 
    else 
        return -1;
}
    
void
_local_del(void *a __attribute__((unused)), void *d __attribute__((unused)))
{
    wrapper_nuke(a);
}

struct wrapper *
wrapper_init(int (*func)(void *, const char *), void *whither)
{
    struct wrapper *retval;
    retval = (struct wrapper *) xcalloc(1, sizeof(struct wrapper));
    retval->func = func;
    retval->where = whither;
    return (retval);
}
    
void
wrapper_nuke(struct wrapper *w)
{
    xfree(w);
}
    


#ifdef TEST

int
main(int argc, char **argv)
{
    read_options("conf");
    return 0;
}
#endif
