/* 
 * Part of the pthread_utils package.  
 *
 * Copyright (c) 2001 Michael Wolf.  All rights reserved.  
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* $Header: /cvsroot/moodns/moodns/src/pool.c,v 1.5 2002/02/01 05:07:51 mawolf Exp $  */

/* A thread in a pool must periodically call pool_check and terminate itself
   if it is asked to.  */

/* pool_init - initialises a pool structure  */

#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>

#include "wpthread.h"
#include "pool.h"
#include "xmalloc.h"

int POOL_MAGICK_DEFAULT = 0xbad5eed; /* ph34r  */

struct pool {
    pthread_mutex_t *mutex;

    int actual_size;

    pthread_t **threads;
    int threads_size;
    
    pthread_t **doomed_threads;
    int doomed_threads_size;
    
    void *(*start_func)(void *);
    void *arg;
};

pool *
pool_init(void *(*func)(void *), void *argument)
{
    pool *retval;

    retval = (pool *) xcalloc(1, sizeof(pool));
    retval->mutex = (pthread_mutex_t *) xmalloc(sizeof(pthread_mutex_t));
    wpthread_mutex_init(retval->mutex, NULL);
    retval->actual_size = 0;
    retval->threads = NULL;  
    retval->threads_size = 0;
    retval->doomed_threads = NULL;
    retval->doomed_threads_size = 0;
    retval->start_func = func;
    retval->arg = argument;
    return (retval);
}

void 
pool_status(pool *p, int *athreads, int *dthreads)
{
    wpthread_mutex_lock(p->mutex);
    if (athreads != NULL)
        *athreads = p->threads_size;
    if (dthreads != NULL) 
        *dthreads = p->doomed_threads_size;
    wpthread_mutex_unlock(p->mutex);
}

int
pool_increment(pool *p)
{
    int retval;
    pthread_t *nt;

    nt = (pthread_t *) xmalloc(sizeof(pthread_t));
    wpthread_create(nt, NULL, p->start_func, p->arg);
    wpthread_mutex_lock(p->mutex);
    ++p->threads_size;
    p->threads = realloc(p->threads, p->threads_size * sizeof(pthread_t *));
    p->threads[p->threads_size - 1] = nt;
    retval = p->threads_size;
    wpthread_mutex_unlock(p->mutex);    

    return (retval);
}

/* FIXME: see the TODO.  */
int
pool_decrement(pool *p)
{
    int retval;
    pthread_t *goner;

    wpthread_mutex_lock(p->mutex);

    /* wpthread_detach(*p->threads[p->threads_size - 1]); */
    goner = p->threads[p->threads_size - 1];
    --p->threads_size;
    p->threads = realloc(p->threads, p->threads_size * sizeof(pthread_t *));
    retval = p->threads_size;

    ++p->doomed_threads_size;
    p->doomed_threads = realloc(p->doomed_threads, p->doomed_threads_size * sizeof(pthread_t *));
    p->doomed_threads[p->doomed_threads_size - 1] = goner;

    wpthread_mutex_unlock(p->mutex);

    return (retval);
}

int
pool_destroy(pool *p __attribute__((unused)))
{
    /* I suppose this ought to be done more or less asynchronously, 
       if possible.  */
    printf("Warning: pool_destroy not implemented yet.\n");
    abort();
    return 0;
}

#define SWAP(n, m) {void *__tmp; __tmp = n; n = m; m = __tmp;}

int
pool_check(pool *p)
{
    int retval;
    pthread_t self;
    int i;

    retval = 0;
    self = wpthread_self();

    wpthread_mutex_lock(p->mutex);
    for (i = 0; i < p->doomed_threads_size; ++i) {
        if (wpthread_equal(self, *p->doomed_threads[i])) {
            SWAP(p->doomed_threads[i], p->doomed_threads[p->doomed_threads_size - 1]);
            pthread_detach(*p->doomed_threads[p->doomed_threads_size - 1]);
            --p->doomed_threads_size;
            p->doomed_threads = realloc(p->doomed_threads, p->doomed_threads_size * sizeof(pthread_t *));
            retval = 1;
            break;
        }
    }
    wpthread_mutex_unlock(p->mutex);

    return (retval);
}
                           



    
    
