/* Copyright (c) 2001, 2002 Omar Kilani, Michael Wolf.  All rights reserved.
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * moodns is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Header: /cvsroot/moodns/moodns/src/protocol-question.c,v 1.4 2002/07/20 17:02:37 mawolf Exp $  */

/* Known issues:
   
   - We don't handle message compression or decompression yet.  

   - Won't compile with -pedantic.  Some of the warnings offered might 
   actually be legitimate.  

 */

/* Based on information presented in RFC 1035, by P Mockapetris.  */

#ifndef _REENTRANT
#define _REENTRANT
#endif

#if HAVE_CONFIG_H
#  include "../config.h"
#endif

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/param.h>  /* FreeBSD needs this for htons etc  */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "record.h"
#include "byte_array.h"
#include "ht.h"
#include "protocol.h"
#include "protocol-question.h"

#include "xmalloc.h"
#if WITH_DMALLOC
#  include <dmalloc.h>
#endif

/* Convert the idxth question section of a packet to raw data.  */

byte_array * 
packet_convert_question_to_data(packet *p, int idx, int *offset, ht *locations)
{
    byte_array *out;
    byte_array *tmp;
    
    out = (byte_array *) xcalloc(1, sizeof(byte_array));

    assert(p->qd != NULL);
    assert(p->qd[idx] != NULL);

    tmp = name_to_bytes(p->qd[idx]->qname, offset, locations);
    byte_array_cat(out, tmp);
    byte_array_catu16(out, htons((short) p->qd[idx]->qtype));
    byte_array_catu16(out, htons((short) p->qd[idx]->qclass));

    byte_array_destroy(tmp);
    
    return out;
}


int
packet_add_question(packet *p, u8 *qname, u32 qtype, u32 qclass)
{
    size_t len;
    struct question *qd;

    qd = (struct question *) xcalloc(1, sizeof(struct question));
    qd->qname = xstrdup(qname);
    qd->qtype = qtype; 
    qd->qclass = qclass; 

    /* stick in into the chain */
    p->h->qdcount++;
    p->qd = (struct question **) 
        xrealloc(p->qd, p->h->qdcount * sizeof(struct question *));
    p->qd[p->h->qdcount - 1] = qd;

    return len = 0; /* XXX: this doesn't mean anything.  make it mean something.  */
}

int
packet_add_question_from_data(packet *p, const u8 *data, int idx, const u8 *base)
{
    const u8 *save;
    u8 *end_qname;
    byte_array *normal = xcalloc(1, sizeof(byte_array));
    size_t len;
    struct question *qd;
    int dsize;

    save = data;

    if (p->h->qdcount <= 0 || idx < 0) {
        fprintf(stderr, "DEBUG: qdcount <= 0 or idx < 0.\n");
        return 0;
    }

    qd = (struct question *) xcalloc(1, sizeof(struct question));
    if ((end_qname = strchr(data, '\0')) == NULL) {
        fprintf(stderr, "DEBUG: No terminator for qname.\n");
        return 0;
    }

    len = end_qname - data;
    qd->qname = question_arr_to_char(data, base, &dsize);
    // p->size_in_bytes += dsize;
    byte_array_cats(normal, qd->qname, strlen(qd->qname) + 1);

    data += len + 1;
    qd->qtype = ntohs(*(const u16 *) & (*data));
    data += 2;
    qd->qclass = ntohs(*(const u16 *) & (*data));
    data += 2;

    /* stick in into the chain, replace whatever is at its index */
    if (p->qd[idx]) {
        if (p->qd[idx]->qname)
            xfree(p->qd[idx]->qname);
        xfree(p->qd[idx]);
    }

    p->qd[idx] = qd;

    return (data - save);

#if 0
    {
        byte_array *compressed;
        byte_array *uncompressed;
        
        /* Get this out of the way.  Will reincorporate it later.  */
        compressed = compress_domain(normal);
        if (compressed != NULL) {
            unsigned int k = 0;
            for (k = 0; k < compressed->len; k++) {
                if (isgraph(compressed->data[k])) {
                    ; 
                } else {
                    ; 
                }
            }
            
            uncompressed = decompress_domain(compressed);
            
            byte_array_destroy(normal);
            byte_array_destroy(compressed);
            byte_array_destroy(uncompressed);
        }
    }
#endif
}


size_t 
question_size(struct question *q)
{
    size_t retval = 0;
    retval += 1 + strlen(q->qname) + 1;
    retval += 2;
    retval += 2;

    return retval;
}


