/* Copyright (c) 2001, 2002 Omar Kilani, Michael Wolf.  All rights reserved.
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * moodns is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Header: /cvsroot/moodns/moodns/src/protocol.c,v 1.23 2002/07/20 17:02:37 mawolf Exp $  */

/* Big picture:

This file implements the dns protocol.  */

/* Known issues:
   
   - question_arr_to_char should take a byte_array as its first arg.  Maybe
   as its second, too.

*/

/* Based on information presented in RFC 1035, by P Mockapetris.  */

#ifndef _REENTRANT
#define _REENTRANT
#endif

#if HAVE_CONFIG_H
#  include "../config.h"
#endif

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/param.h>  /* FreeBSD needs this for htons etc  */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <limits.h>

#include "record.h"
#include "ht.h"
#include "byte_array.h"
#include "protocol.h"
#include "protocol-header.h"
#include "protocol-question.h"
#include "protocol-rr.h"

#include "xmalloc.h"
#if WITH_DMALLOC
#  include <dmalloc.h>
#endif


typedef struct {
    u8 **found;
    u32 *location;
    u32 n;
} compress_table;

static int _domain_compare(void *, void *, void *);
static void _domain_delete(void *, void *);

static int _local_cmp(void *, void *, void *);
static void _local_del(void *, void *);

/* Construct an empty packet structure.  */

packet *
packet_create()
{
    packet *p = xcalloc(1, sizeof(packet));
    return p;
}

/* Construct a packet from raw data.  */

packet * 
packet_create_from_data(const byte_array *ba)
{   /* Was (const u8 *data, u32 len)  */
    packet *p;
    struct header_packed hp;
    const u8 *section; /* Where we are within data.  */
    size_t ssize;
    int i;
    
    if (ba->data == NULL || ba->len < sizeof(hp.data)) {
        fprintf(stderr, "DEBUG: malformed query received.\n"); 
        return NULL;
    }
    
    /* allocate packet structure  */
    p = (packet *) xcalloc(1, sizeof(packet));
    
    /* allocate packet's header  */
    p->h = (struct header *) xcalloc(1, sizeof(struct header));
    
    /* place a packet header's size worth of bytes to the packed header  */
    memset(&hp, 0, sizeof(hp));
    memcpy(hp.data, ba->data, sizeof(hp.data));
    
    /* unpack those bytes into header structure  */
    header_unpack(p->h, &hp);
    p->size_in_bytes += sizeof(hp.data);
    
    section = ba->data + sizeof(hp.data);   
    
    /* allocate question  */
    if (p->h->qdcount != 0)
        p->qd = (struct question **) 
            xcalloc(p->h->qdcount, sizeof(struct question *));
    /* convert questions from raw bytes into question structures, place
       question structures into packet  */
    for (i = 0; i < p->h->qdcount; i++) {
        ssize = packet_add_question_from_data(p, section, i, ba->data);
        p->size_in_bytes += ssize;
        section += ssize;
    }

   /* convert answers from raw bytes into rr structures, put
      structures into packet  */
    if (p->h->ancount != 0) 
        p->an = (struct resource_record **)
            xcalloc(p->h->ancount, sizeof(struct resorce_record *));
    for (i = 0; i < p->h->ancount; i++) {
        ssize = packet_add_answer_from_data(p, section, i, ba->data);
        p->size_in_bytes += ssize;
        section += ssize;
    }
        
    /* convert authorities from raw bytes into rr structures, put
       structures into packet  */
    if (p->h->nscount != 0)
        p->ns = (struct resource_record **) 
            xcalloc(p->h->nscount, sizeof(struct resource_record *));
    for (i = 0; i < p->h->nscount; i++) {
        ssize = packet_add_authoritive_from_data(p, section, i, ba->data);
        p->size_in_bytes += ssize;
        section += ssize;
    }

    /* convert additionals from raw bytes into rr structures, put
       structures into packet  */ 
    if (p->h->arcount != 0)
        p->ar = (struct resource_record **)
            xcalloc(p->h->arcount, sizeof(struct resource_record *));
    for (i = 0; i < p->h->arcount; i++) {
        ssize = packet_add_additional_from_data(p, section, i, ba->data);
        p->size_in_bytes += ssize;
        section += ssize;
    }

    return p;
}

/* Convert a packet to data suitable to sending over the net.  */

byte_array * 
packet_convert_to_data(packet *p)
{
    byte_array *out;
    byte_array *section;
    struct header_packed hp;
    int i;
    int offset;
    ht *locations;
    
    offset = 0;
    locations = ht_init(_local_cmp, _local_del);

    out = (byte_array *) xcalloc(1, sizeof(byte_array));
    header_pack(p->h, &hp);
    byte_array_cats(out, (const u8 *) hp.data, sizeof(hp.data));

    for (i = 0; i < p->h->qdcount; i++) {
        offset = out->len;
        section = packet_convert_question_to_data(p, i, &offset, locations);
        byte_array_cat(out, section);
        byte_array_destroy(section);
    }

    for (i = 0; i < p->h->ancount; i++) {
        section = packet_convert_answer_to_data(p, i, &offset, locations);
        byte_array_cat(out, section);
        byte_array_destroy(section);
    }

    for (i = 0; i < p->h->nscount; i++) {
        section = packet_convert_authoritive_to_data(p, i, &offset, locations);
        byte_array_cat(out, section);
        byte_array_destroy(section);
    }

    for (i = 0; i < p->h->arcount; i++) {
        section = packet_convert_additional_to_data(p, i, &offset, locations);
        byte_array_cat(out, section);
        byte_array_destroy(section);
    }

    return out;
}

/* Trivially print out the contents of a packet.  */

void
packet_debug_dump(packet *p)
{
    printf("printing packet's header:\n");
    printf("+ id = %d\n+ qr = %d\n+ opcode = %d\n+ aa = %d\n"
           "+ tc = %d\n+ rd = %d\n", p->h->id, p->h->qr, p->h->opcode, 
           p->h->aa, p->h->tc, p->h->rd);
    printf("+ ra = %d\n+ z = %d\n+ rcode = %d\n+ qdcount = %d\n"
           "+ ancount = %d\n", p->h->ra, p->h->z, p->h->rcode, p->h->qdcount,
           p->h->ancount);
    printf("+ nscount = %d\n+ arcount = %d\n", p->h->nscount, p->h->arcount);
    
    {
        int iqd;
        printf("printing packet's query sections:\n");
        for (iqd = 0; iqd < p->h->qdcount; ++iqd) {
            printf("+ query name: %s\n", p->qd[iqd]->qname);
            assert(p->qd[iqd]->qclass == IN); /* For now, anyway.  */
            printf("+ query type: %d (%s)\n", p->qd[iqd]->qtype, 
                   record_type_names[p->qd[iqd]->qtype]);
            printf("+ query class: %d (%s)\n", p->qd[iqd]->qclass,
                   record_class_names[p->qd[iqd]->qclass]);
        }
    }
    
    {
        int ian;
        printf("there are %d records in the answer section\n", p->h->ancount);
        for (ian = 0; ian < p->h->ancount; ++ian) {
            struct in_addr ia;
            printf("answer %d:\n", ian);
            printf("    name = %s\n", p->an[ian]->name);
            printf("    type = %d (%s)\n", p->an[ian]->type, 
                   record_type_names[p->an[ian]->type]);
            printf("    class = %d (%s)\n", p->an[ian]->class,
                   record_class_names[p->an[ian]->class]);
            printf("    ttl = %u\n", p->an[ian]->ttl);
            
            if (p->an[ian]->type == A) {
                ia.s_addr = htonl(*(int *) p->an[ian]->rdata->data);
                printf("    rdata is %s\n", inet_ntoa(ia));
            } else {
                printf("    sorry, can't print that kind of rdata yet :(\n");
            }
        }
    }
    {
        int ins;
        printf("there are %d records in the ns section\n", p->h->nscount);
        for (ins = 0; ins < p->h->nscount; ++ins) {
            struct in_addr ia;
            printf("answer %d:\n", ins);
            
            printf("    name = %s\n", p->ns[ins]->name);
            printf("    type = %d (%s)\n", p->ns[ins]->type, 
                   record_type_names[p->ns[ins]->type]);
            printf("    class = %d (%s)\n", p->ns[ins]->class,
                   record_class_names[p->ns[ins]->class]);
            printf("    ttl = %u\n", p->ns[ins]->ttl);
            
            if (p->ns[ins]->type == A) {
                ia.s_addr = htonl(*(int *) p->ns[ins]->rdata->data);
                printf("    rdata is %s\n", inet_ntoa(ia));
            } else {
                printf("    sorry, can't print that kind of rdata yet :(\n");
            }   ;
        }
    }
    
    {
        int iar;
        printf("there are %d records in the authoritative section\n", p->h->arcount);
        for (iar = 0; iar < p->h->arcount; ++iar) {
            struct in_addr ia;
            printf("answer %d:\n", iar);
            printf("    name = %s\n", p->ar[iar]->name);
            printf("    type = %d (%s)\n", p->ar[iar]->type, 
                   record_type_names[p->ar[iar]->type]);
            printf("    class = %d (%s)\n", p->ar[iar]->class,
                   record_class_names[p->ar[iar]->class]);
            printf("    ttl = %u\n", p->ar[iar]->ttl);
            
            if (p->ar[iar]->type == A) {
                ia.s_addr = htonl(*(int *) p->ar[iar]->rdata->data);
                printf("    rdata is %s\n", inet_ntoa(ia));
            } else {
                printf("    sorry, can't print that kind of rdata yet :(\n");
            };
        }
    }
    printf("In serialised form, this packet is %u bytes long.\n",
           p->size_in_bytes);
}

void
packet_free(packet *p)
{
    int i;

    if (p->qd) {
        for (i = 0; i < p->h->qdcount; i++) {
            if (p->qd[i] && p->qd[i]->qname)
                xfree(p->qd[i]->qname);
            if (p->qd[i])
                xfree(p->qd[i]);
        }
    }

    if (p->an) {
        for (i = 0; i < p->h->ancount; i++) {
            xfree(p->an[i]);
        }
    }

    if (p->ns) {
        for (i = 0; i < p->h->nscount; i++) {
            xfree(p->ns[i]);
        }
    }

    if (p->ar) {
        for (i = 0; i < p->h->arcount; i++) {
            xfree(p->ar[i]);
        }
    }
    
    if (p->h)
        xfree(p->h);

    if (p)
        xfree(p);
}


/* Convert a "question array" to an standard '\0'-delimited C string.
   bytes is the raw data we need to convert.  When a compression
   pointer is found, the offset is with respect to base.  dsize is
   set to the correct size to be added to a size_in_bytes sum, which
   will be less than the size of the question array if compression is
   in fact used.  

   Use question_arr_to_char_nd when you're sure there will be no
   decompression.  */

char *
question_arr_to_char(const u8 *bytes, const u8 *base, int *dsize)
{
    char *retval;
    int x, y, z;
    const u8 *c;
    u16 base_offset;
    int keep_adding = 1; /* Gross.  */
    int first_part = 1;

    assert(bytes != NULL);
    assert(dsize != NULL);
    
    *dsize = 0;
    retval = xcalloc(256, sizeof(char));
    c = bytes;

    x = *c;
    z = 0;
    do {
        /* Check for compression pointer - the first 2 bits will be set  */
        if (*c & (128 | 64)) { 
            puts("compression pointer found");
            /* This is slightly risky, but the assert should catch problems
               if they crop up.  */
            assert(base != NULL);  
            /* XXX: Potentially dangerous in the face of bogus input?  Change
               first argument (at least) to a byte array so we can know for
               sure.  */
            base_offset = c[1]; 
            c = &base[base_offset];
            x = *c;
            if (keep_adding)
                *dsize += 2;
            keep_adding = 0;
        }
        if (!isalpha(*c) && first_part) {
            first_part = 0;
            if (keep_adding)
                ++*dsize;
        }
        for (y = 0; y < x; ++y) {
            retval[z++] = *(c + 1 + y);
            if (keep_adding)
                ++*dsize;
        }
        c += (x + 1);
        x = *c;
        if (x == '\0') {
            if (keep_adding) /* XXXXX: May be wrong.  */
                ++*dsize;
            break;
        }
        retval[z++] = '.';
        if (keep_adding)
            ++*dsize;
    } while (x != 0);
    return retval;
}

char *
question_arr_to_char_nd(const u8 *bytes, int *dsize)
{
    char *retval;
    int x, y, z;
    const u8 *c;
    int first_part;
    
    assert(bytes != NULL);
    assert(dsize != NULL);
    
    first_part = 1;
    *dsize = 0;
    retval = xcalloc(256, sizeof(char));
    c = bytes;
    x = *c;
    z = 0;
    do {
        if (!isalpha(*c) && first_part) {
            first_part = 0;
            ++*dsize;
        }
        for (y = 0; y < x; ++y) {
            retval[z++] = *(c + 1 + y);
            ++*dsize;
        }
        c += (x + 1);
        x = *c;
        if (x == '\0') {
            ++*dsize;
            break;
        }
        retval[z++] = '.';
        ++*dsize;
    } while (x != 0);
    return retval;
}

byte_array *
name_to_bytes(const char *name, int *offset, ht *locations)
{
    byte_array *retval;
    char *namecpy;
    char *p, *q, r;
    u8 d;
    char *tmp;
    int off;
    u16 ioff;

    retval = xcalloc(1, sizeof(byte_array));
    namecpy = alloca(strlen(name) + 1);
    strcpy(namecpy, name);
    p = namecpy;
    off = *offset;
    // while (*p != '\0') {
    do {
        /* We're going to cheat and assume that pointers and integers are
           equivalent here.  If that actually fails on some platform, ifdef 
           in something slower.  */
        if (( ioff = (int) ht_get(locations, p)) == 0)
            ht_put(locations, p, (void *) off, NULL);
        else {
            ioff &= (128 | 64);
            byte_array_catu16(retval, ioff);
            break;
        }
        q = p;
        while (*q != '.' && *q != '\0')
            ++q;
        d = q - p;
        r = *q;
        *q = '\0';
        /* 1 for the length byte, d for the chars, 1 for the '\0'
           which is needed with strncpy.  I think.  */
        tmp = (char *) alloca(1 + d + 1);
        memset(tmp, '\0', d + 2);
        tmp[0] = d;
        strncpy(&tmp[1], p, d);
        byte_array_cats(retval, tmp, d + 1);
        p = q + 1;
    } while (r != '\0');
    byte_array_catu8(retval, '\0'); 
    return retval;
}

/* Take a potentially compressed domain name and convert it to
   uncompressed form.  */

byte_array *
decompress2(const u8 *pc, const u8 *base)
{
    byte_array *retval;
    int x, y, z;
    const u8 *c;
    u16 base_offset;

    retval = (byte_array *) xcalloc(1, sizeof(byte_array));
    assert(pc != NULL);
    c = pc;
    x = *c;
    z = 0;
    do {
        if (*c & (128 | 64)) {
            puts("compression pointer found");
            base_offset = c[1];
            c = &base[base_offset];
            x = *c;
        }
        byte_array_catu8(retval, x);
        for (y = 0; y < x; ++y) {
            byte_array_catu8(retval, *(c + 1 + y));
        }
        c += (x + 1);
        x = *c;
        if (x == '\0') 
            break;
        // byte_array_catu8(retval, '.');


    } while (x != 0);
    byte_array_catu8(retval, '\0');
    return retval;
}

int
_domain_compare(void *a, void *b, void *aux __attribute__((unused)))
{
    if (strcmp((u8 *)a, (u8 *)b) == 0)
        return 0;
    else
        return 1;
}

void
_domain_delete(void *a, void *aux __attribute__((unused)))
{
    xfree((u8 *)a);
}

byte_array *
compress_domain(byte_array *decompressed)
{
    byte_array *compressed = (byte_array *) xcalloc(1, sizeof(byte_array));
    u8 *dataptr, *l = NULL;
    ht *h = ht_init(_domain_compare, _domain_delete);
    u32 *location = NULL;
   
    byte_array_cat(compressed, decompressed);
    dataptr = compressed->data;
 
    while (*dataptr != 0) {
        u32 len;

        if (*dataptr > 63 || *dataptr <= 0)
            return NULL;

        len = *dataptr;
        l = (u8 *) realloc(l, len + 1);
        strncpy(l, dataptr + 1, len);
        l[len] = 0;
        if ((location = (u32 *) ht_get(h, l)) != NULL) {
            u8 pointer[1];
            u32 offset = dataptr - compressed->data;
            
            byte_array_delete(compressed, offset, len+1); 
            *pointer = 0;
            *(pointer + 1) = 0;
            *pointer |= (*location << 8);
            *pointer |= 192; 
            *(pointer+1) |= (*location);
            byte_array_inserts(compressed, offset, pointer, 2);

            dataptr = compressed->data;
            dataptr += offset + sizeof(u16);
        } else {
            location = (u32 *) xmalloc(sizeof(u32));
            *location = dataptr - compressed->data;
            ht_put(h, l, location, NULL);
            dataptr += strlen(l) + 1;
        }

        xfree(l);
        l = NULL;
    }

    return compressed;
}

byte_array *
decompress_domain(byte_array *compressed)
{
    byte_array *decompressed = (byte_array *) xcalloc(1, sizeof(byte_array));
    u8 *dataptr;
    
    byte_array_cat(decompressed, compressed);
    dataptr = decompressed->data + decompressed->len;
    
    while (dataptr != decompressed->data) {
        if (*dataptr > 63) {
            if (*dataptr < 192) {
                /* EDNS ? */
                dataptr--;
            } else {
                u8 offset, len, loc;
                u8 *label;

                offset = ((*dataptr & 0x3F) << 8) | *(dataptr + 1);
                len = *(decompressed->data + offset);
                loc = dataptr - decompressed->data;

                label = (u8 *) xmalloc(len+1);
                strncpy(label, decompressed->data + offset + 1, len);
                label[len] = 0;
                
                byte_array_delete(decompressed, loc, 2);
                byte_array_inserts(decompressed, loc, &len, 1);
                byte_array_inserts(decompressed, loc + 1, label, strlen(label));
                dataptr = decompressed->data;
                dataptr += loc - 1;

                xfree(label);
                label = NULL; 
            }
        } else {
            dataptr--;
        }
    }

    return decompressed; 
}



/* These really belong in misc.c  */

void
bitprint_u8(u8 c)
{
    int i = UCHAR_MAX - 1;
    while (i > 0) {
        if (i & c) 
            fputc('1', stdout);
        else 
            fputc('0', stdout);
        i >>= 1;
        fputc(' ', stdout);
    }
    puts("");
}

void
bitprint_u16(u16 s)
{
    int i = USHRT_MAX;
    while (i > 0) {
        if (i & s) 
            fputc('1', stdout);
        else
            fputc('0', stdout);
        i >>= 1;
    } 
    puts("");
}

void
bitprint_u32(u32 l)
{
    unsigned int i = UINT_MAX - 1;
    while (i > 0) {
        if (i & l)
            fputc('1', stdout);
        else
            fputc('0', stdout);
        i >>= 1;
    }
    puts("");
}





#ifdef TEST

#include <stdio.h>
#include <stdlib.h>

int main(int argc __attribute__((unused)), char **argv __attribute__((unused)))
{

    struct header h, h2;
    struct header_packed p;
    packet *x, *x2;

    memset(&h, 0, sizeof(h));

    h.id = 12345;
    h.qr = 0;
    h.opcode = 1;
    h.aa = 0;
    h.tc = 1;
    h.rd = 1;
    h.ra = 2;
    h.z = 0;
    h.rcode = 2;
    h.qdcount = 1;
    h.ancount = 5;
    h.nscount = 0;
    h.arcount = 7;

    header_pack(&h, &p);
    header_unpack(&h2, &p);

    assert(h.id == h2.id);
    assert(h.qr == h2.qr);
    assert(h.opcode == h2.opcode);
    assert(h.aa == h2.aa);
    assert(h.tc == h2.tc);
    assert(h.rd == h2.rd);
    assert(h.ra == h2.ra);
    assert(h.z == h2.z || h2.z == 0);   /* Good test?  */
    assert(h.rcode == h2.rcode);
    assert(h.qdcount == h2.qdcount);
    assert(h.ancount == h2.ancount);
    assert(h.nscount == h2.nscount);
    assert(h.arcount == h2.arcount);

    /* This will fail next time we attempt to compile the tests.  */
    x = packet_create();
    x2 = packet_create();

    


    packet_free(x);
    packet_free(x2);
    return 0;
}

#endif



int
_local_cmp(void *a, void *b, void *c __attribute__((unused))) 
{
    return strcmp((const char *) a, (const char *) b);
}

void
_local_del(void *a, void *b __attribute__((unused)))
{
    xfree(a);
}


