/* Copyright (c) 2001 Omar Kilani, Michael Wolf.  All rights reserved.
 * Copyright (c) 2002 Omar Kilani, Michael Wolf.  All rights reserved.
 * Copyright (c) 2002 Ximian, Inc.  All rights reserved.
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * moodns is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Header: /cvsroot/moodns/moodns/src/protocol.h,v 1.17 2002/07/20 17:02:37 mawolf Exp $  */

#ifndef PROTOCOL_H
#define PROTOCOL_H

#include "ht.h"
#include "record.h"
#include "byte_array.h"

enum QTYPE { AXFR = 252, MAILB = 253, MAILA = 254, STAR = 255 };

struct packet {
    struct header *h;
    struct question **qd;
    struct resource_record **an;
    struct resource_record **ns;
    struct resource_record **ar;
    /* The size we're concerned with is the packed size.  We don't
       really care how large it is unpacked.  Necessary (at least) so
       we know when to set the tc bit in the header.  Also should be 
       useful while debugging.  */
    size_t size_in_bytes;
};


struct header {
    /* Unique (from the client) identifier.  */
    u32 id:16;
    
    /* 0 if a query, 1 if a response.  */
    u32 qr:1;
    
    /* Only 4 bits used.  0: std query.  1: inverse query.  2: server status
       request.  3-15: reserved for future use.  Must be less than 16, and 
       generally less than 3.  */
    u32 opcode:4;
    
    /* Authoritative answer.  */
    u32 aa:1;
    
    /* Truncation: true if message had to be truncated.  */
    u32 tc:1;
    
    /* Recursion Desired: true if we want server to do a recursive search.  */
    u32 rd:1;
    
    /* Recursion Available: true if the server supports it, else false.  */
    u32 ra:1;
    
    /* Reserved for future use.  Must always be zero.  3 bits.  */
    u32 z:3;
    
    /* response code.  0: no error condition.  1: format error.  2: server
       failure.  3: name error.  4: not implemented.  5: refused (for policy
       reasons).  6 - 15: reserved for future use.  */
    u32 rcode:4;
    
    /* Number of entries in the question section.  */
    u32 qdcount:16;
    
    /* Number of resource records in the answer section.  */
    u32 ancount:16;
    
    /* Number of name server resource records in the authority 
       records section  */
    u32 nscount:16;
    
    /* Number of resource records in the additional records section.  */
    u32 arcount:16;
};

struct header_packed {
    byte_array *data;
};


struct question {
    char *qname;
    /* All codes valid for a TYPE field, plus some general codes for more than
       one type of RR.  */
    u32 qtype:16;
    /* IN for internet.  Could conceivably be something else...?  */
    u32 qclass:16;
    
    size_t size_in_bytes;
};

struct resource_record {
    char *name;
    u32 type:16;
    u32 class:16;
    u32 ttl;
    u16 rdlength; /* Why wasn't this here before?  */
    byte_array *rdata;
    /* This won't be compressed, and isn't included in any size_in_bytes
       calculations.*/
    byte_array *rdata_uncompressed;

    size_t size_in_bytes;
};


typedef struct packet packet;


packet *packet_create(void);
packet *packet_create_from_data(const byte_array *);
// packet *packet_create(u32 id, u32 qr, u32 opcode, u32 aa, u32 tc, u32
// rd, u32 ra, u32 z, u32 rcode); 
int packet_add_header(u32, u32, u32, u32, u32, u32, u32, u32, u32);
int packet_add_question(packet *, u8 *, u32, u32);
int packet_add_rr(packet *p, u8 *name, u32 type, u32 class, u32 tll,
                  u32 rdlength, u8 *rdata);
void packet_debug_dump(packet *);
void packet_free(packet *);
byte_array *packet_convert_to_data(packet *);


/* FIXME: These are bad names.  Think of better ones.  */
char *question_arr_to_char(const u8 *, const u8 *, int *);
char *question_arr_to_char_nd(const u8 *, int *);
byte_array *name_to_bytes(const char *, int *, ht *);
byte_array *decompress2(const u8 *, const u8 */*, u16*/);

byte_array *compress_domain(byte_array *);
byte_array *decompress_domain(byte_array *);


// int query_parse(char *data, int len);

struct resource_record *record_to_resource_record(record *, int *, ht *);
record *resource_record_to_record(struct resource_record *);

size_t header_size(struct header *);
size_t question_size(struct question *);
size_t resource_record_size(struct resource_record *);


/* *** This is the new API. ***  */
packet *packet_init(void);
enum query_response { QUERY = 0, RESPONSE, NOCHANGE_QUERY_RESPONSE };
enum query_response packet_query_response(packet *, enum query_response);

enum opcode { STD = 0, INVERSE, STATUS, NOCHANGE_OPCODE };
enum opcode packet_opcode(packet *, enum opcode);

enum header_bool { NO_HEADER_BOOL, YES_HEADER_BOOL, NOCHANGE_HEADER_BOOL };
enum header_bool packet_authoritative(packet *, enum header_bool);


#endif
