/* Copyright (c) 2001, 2002 Michael Wolf.  All rights reserved. 
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * moodns is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
#include <string.h>

#include "record.h"

#include "xmalloc.h"
#if WITH_DMALLOC
# include <dmalloc.h>
#endif

const char *record_class_names[] = {
    "BOGUS",
    "IN - Internet",
    "CS - CSNET class (obsolete)",
    "CH - CHAOS",
    "HS - Hesiod",
    NULL
};


const char *record_type_names[] = {
    "BOGUS", 
    "A", 
    "NS", 
    "MD", 
    "MF", 
    "CNAME", 
    "SOA", 
    "MB", 
    "MG", 
    "MR", 
    "NULL_EXP", 
    "WKS",
    "PTR", 
    "HINFO", 
    "MINFO",
    "MX", 
    "TXT", 
    NULL 
};

record *
record_alloc()
{
    /* XXX: Consider replacing this with a custom allocator.  (Pre-allocate 
       lots of records and dole them out and reclaim them as necessary.  That 
       might avoid heap fragmentation and general expensiveness of malloc.)  */
    return (record *) xcalloc(1, sizeof(record));
}

record *
record_alloc_dup(record *r)
{
    record *retval = xmalloc(sizeof(*retval));
    memmove(retval, r, sizeof(*retval));
    return retval;
}

/* return 0 if they're the same, nonzero if they're different.  */

int 
record_compare(record *a, record *b)
{
    int retval = 1;

    if (strcmp(a->name, b->name) != 0)
        goto done;
    if (a->ip != b->ip)
        goto done;
    if (a->expiry != b->expiry)
        goto done;

    retval = 0;
done:
    return (retval);
}

/* return 0 if they're the same, nonzero if they're different.  */

int
record_ttl_compare(record *a, record *b)
{
    if (a->expiry >= b->expiry)
        return 1;
    else
        return 0;
}

unsigned int
record_expiry(record *a)
{
    return (a->expiry);
}

void
record_delete(record *a)
{
    /* XXX: see comments in record_alloc.  */
    xfree(a);
}

void
record_debug_dump(record *a)
{
    printf("printing record:\n");
    printf(" type = %s\n", record_type_names[a->type]);
    printf(" name = %s\n", a->name);
    switch (a->type) {
    case A:
        printf(" A = %u\n", a->record.a); 
        break;
    case NS:
        printf(" NS = %s\n", a->record.ns);
        break;
    case CNAME:
        printf(" CNAME = %s\n", a->record.cname);
        break;
    case SOA:
        printf(" SOA = ...\n");
        break;
    case PTR: 
        printf(" PTR = %s\n", a->record.ptr);
        break;
    case MX:
        printf(" MX = %d %s\n", a->record.mx.preference,
               a->record.mx.exchange);
        break;
    case TXT:
        printf(" TXT = '%s'\n", a->record.txt);
        break;
    default:
        break;
    }
    printf(" expiry = %u\n", a->expiry);
    printf(" ttl = %d\n", a->ttl);
    printf(" timestamp = %qu\n", a->timestamp);
}

struct record_chain *
record_chain_init(char *name)
{
    struct record_chain *retval;
    retval = (struct record_chain *) xcalloc(1, sizeof(struct record_chain));
    retval->name = xstrdup(name);
    return (retval);
}

int
record_chain_add(struct record_chain *c, struct record *r)
{
    int retval;
    
    c->records = xrealloc(c->records, ++c->records_len * sizeof(record));
    /* Can this fail?  */
    memmove(&c->records[c->records_len - 1], r, sizeof(record));
    
    return retval = 0;
}

int
record_chain_add_chain(struct record_chain *c, struct record_chain *d)
{
    int retval;
    size_t old;
    old = c->records_len * sizeof(record);
    c->records_len += d->records_len;
    c->records = xrealloc(c->records, c->records_len * sizeof(record));
    memmove(&c->records[old], d->records, d->records_len * sizeof(record));
    return retval = 0;
}

struct record **
record_chain_get(struct record_chain *c, enum record_type type)
{
    struct record **retval = NULL;
    int retval_len = 0;
    int i;

    for (i = 0; i < c->records_len; ++i) {
        if (c->records[i].type == type) {
            retval = (struct record **) 
                xrealloc(retval, ++retval_len * sizeof(struct record *));
            retval[i - 1] = &c->records[i];
        }
    }
    if (retval_len > 0) {
        retval = (struct record **) 
            xrealloc(retval, ++retval_len * sizeof(struct record *));
        retval[retval_len - 1] = NULL;
    }
    return (retval);
}

void
record_chain_destroy(struct record_chain *rc)
{
    int i;
    
    /* records, rather than pointers to records, are copied directly into 
       record chains, so we delete no records...  */
    for (i = 0; i < rc->records_len; ++i) {
        ;
    }
    /* ... but we do free the chain itself ...  */
    xfree(rc->records);
    /* ... and its name.  */
    xfree(rc->name);
}
