/* Copyright (c) 2001 Omar Kilani.  All rights reserved. 
 * Copyright (c) 2002 Michael Wolf.  All rights reserved.
 *
 * This file is part of moodns.
 *
 * moodns is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 * 
 * moodns is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with moodns; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _REENTRANT
#define _REENTRANT
#endif

#if HAVE_CONFIG_H
  #include "../config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <assert.h>

#include "util.h"
#include "byte_array.h"
#include "mtime.h"
#include "socket.h"

#include "xmalloc.h"
#if WITH_DMALLOC
  #include <dmalloc.h>
#endif

static void socket_set_minimum(socket_t *, size_t);
static int socket_seqpacket_handshake(socket_t *);
static int handshake_send(const char *, socket_t *);
static int handshake_recv(char *, const char *, size_t, socket_t *s);

/* Handy shorthand.  */

typedef struct sockaddr sa;

/* Create a listening socket.  */

socket_t *
socket_listener(socket_address_t *addr, enum socket_domain domain, 
                enum socket_type type)
{
    socket_t *s = (socket_t *) xcalloc(1, sizeof(socket_t));
    int reuse = -1;
    int ret;
    int sdomain, stype;

    if (!s)
        return NULL;

    if (domain == SOCKET_INTERNET)
        sdomain = PF_INET;
    else if (domain == SOCKET_LOCAL)
        sdomain = PF_LOCAL;
    else
        abort();
    
    if (type == SOCKET_STREAM)
        stype = SOCK_STREAM;
    else if (type == SOCKET_DATAGRAM)
        stype = SOCK_DGRAM;
    else if (type == SOCKET_SEQPACKET)
        stype = SOCK_STREAM;  
    else
        abort();
    
    if ((s->fd = socket(sdomain, stype, 0)) == -1) {
        perror("socket failed");
        xfree(s);
        return NULL;
    }

    if (setsockopt(s->fd, SOL_SOCKET, SO_REUSEADDR, &reuse, (socklen_t) sizeof(reuse)) < 0)
        fprintf(stderr, "setsockopt error.\n");

    if (domain == SOCKET_INTERNET) {
        s->socket_address.internet.sin_family = PF_INET;
        s->socket_address.internet.sin_port = htons(addr->internet.port);
        if (addr->internet.address == NULL) {
            s->socket_address.internet.sin_addr.s_addr = INADDR_ANY;
        } else {
            // Fix this.
            // s->socket_address.internet.sin_addr.s_addr = something;
            s->socket_address.internet.sin_addr.s_addr = INADDR_ANY;

        }
    } else if (domain == SOCKET_LOCAL) {
        s->socket_address.local.sun_family = PF_LOCAL;
        strcpy(s->socket_address.local.sun_path, addr->local.path);
    } else {
        abort();
    }

    if (domain == SOCKET_INTERNET)
        ret = bind(s->fd, (struct sockaddr *) &s->socket_address.internet, 
                   (socklen_t) sizeof(struct sockaddr_in));
    else if (domain == SOCKET_LOCAL)
        ret = bind(s->fd, (struct sockaddr *) &s->socket_address.local,
                   (socklen_t) sizeof(struct sockaddr_un));
    else
        abort();
         
    if (ret == -1) {
        perror("bind failed");
        ret = close(s->fd);
        if (ret == -1)
            perror("close failed too");
        xfree(s);
        return NULL;
    }

    if ((type == SOCKET_STREAM || type == SOCKET_SEQPACKET)
        && listen(s->fd, 40) == -1) {
        perror("listen failed");
        ret = close(s->fd);
        if (ret == -1)
            perror("close failed too");
        xfree(s);
        return NULL;
    }

    s->domain = domain;
    s->type = type;
    s->job = SOCKET_LISTENER;

    return s;
}

/* Create a sending socket.  */

socket_t *
socket_sender(socket_address_t *addr, enum socket_domain domain, 
              enum socket_type type)
{
    socket_t *s = (socket_t *) xcalloc(1, sizeof(socket_t));
    int sdomain, stype;
    int ret;

    if (domain == SOCKET_INTERNET)
        sdomain = PF_INET;
    else if (domain == SOCKET_LOCAL)
        sdomain = PF_LOCAL;
    else
        abort();

    if (type == SOCKET_STREAM)
        stype = SOCK_STREAM;
    else if (type == SOCKET_DATAGRAM)
        stype = SOCK_DGRAM;
    else if (type == SOCKET_SEQPACKET)
        stype = SOCK_STREAM;
    else
        abort();

    if ((s->fd = socket(sdomain, stype, 0)) == -1) {
        perror("socket failed");
        xfree(s);
        return NULL;
    }

    if (domain == SOCKET_INTERNET) {
        struct in_addr in;
        u32 remote;
        if (addr->internet.address != NULL) {
            if (inet_aton(addr->internet.address, /*@out@*/ &in) == 0) {
                xfree(s);
                return NULL;
            }
        } else {
            in.s_addr = addr->internet.ip;
        }
        remote = (u32) in.s_addr;         
        s->socket_address.internet.sin_family = PF_INET;
        s->socket_address.internet.sin_port = htons(addr->internet.port);
        s->socket_address.internet.sin_addr.s_addr = remote;
    } else if (domain == SOCKET_LOCAL) {
        s->socket_address.local.sun_family = PF_LOCAL;
        strcpy(s->socket_address.local.sun_path, addr->local.path);
    } else {
        abort();
    }

    if (domain == SOCKET_INTERNET) {
        ret = connect(s->fd, (struct sockaddr *) &s->socket_address.internet, 
                      sizeof(struct sockaddr_in));
    } else if (domain == SOCKET_LOCAL) {
        ret = connect(s->fd, (struct sockaddr *) &s->socket_address.local, 
                      sizeof(struct sockaddr_un));
    } else {
        abort();
    }

    if (ret == -1) {
        perror("connect failed");
        xfree(s);
        return NULL;
    }

    s->domain = domain;
    s->type = type;
    s->job = SOCKET_SENDER;

    if (s->type == SOCKET_SEQPACKET) {
        ret = socket_seqpacket_handshake(s);
        if (ret == 0) {
            /* XXX: Clean up.  */
            goto done;
        }
    }

done:
    return s;
}

socket_t *
socket_accept(socket_t *s)
{
    socket_t *client = (socket_t *) xcalloc(1, sizeof(socket_t));
    
    if (!client)
        return NULL;
    client->domain = s->domain;
    client->type = s->type;
    client->job = SOCKET_CLIENT;

    if (s->type == SOCKET_STREAM) {
        size_t size;

        if (s->domain == SOCKET_INTERNET) 
            size = sizeof(struct sockaddr_in);
        else if (s->domain == SOCKET_LOCAL)
            size = sizeof(struct sockaddr_un);

        client->fd = accept(s->fd, (sa *) &client->socket_address.internet,
                            &size);
    } else if (s->type == SOCKET_DATAGRAM) {
        client->fd = s->fd;
    } else if (s->type == SOCKET_SEQPACKET) {
        size_t size;
        int h;
        client->fd = accept(s->fd, (sa *) &client->socket_address.internet,
                            &size);
        h = socket_seqpacket_handshake(client);
        /* XXX: check h  */
    } else {
        abort();
    }
    
    return client;
}

/* close a socket  */

int
socket_close(socket_t *s, enum socket_close how)
{
    /* XXX: fix this  */
    int ret;
    assert(how == SOCKET_READING || how == SOCKET_WRITING || 
           how == SOCKET_BOTH);
    ret = shutdown(s->fd, how);
    if (ret == -1 && errno != ENOTCONN) {
        perror("shutdown failed");
        abort();
    }
    if (s->domain == SOCKET_LOCAL && s->job == SOCKET_LISTENER) {
        ret = unlink(s->socket_address.local.sun_path);
        if (ret == -1)
            abort(); /* XXX: not sure if this is correct behaviour.  */
    }

    xfree(s);
    return 1;
}

/* Receive data from a socket.  The data that is received is appended to the 
   provided byte_array.  (I think that's the most useful.  Do you?)  A maximum
   of maxlen bytes will be read.  If maxlen is zero, then as many bytes as are
   available will be read.  */

/* Block indefinitely.  */

size_t 
socket_recv(socket_t *s, byte_array *data, size_t maxlen)
{
    return socket_recv_timeout(s, data, maxlen, -1, -1);
}

/* Don't block.  */

size_t 
socket_recv_poll(socket_t *s, byte_array *data, size_t maxlen)
{
    return socket_recv_timeout(s, data, maxlen, 0, 0);
}

/* Block for the specified amount of time.  */

size_t
socket_recv_timeout(socket_t *s, byte_array *data, size_t maxlen, 
                    int secs, int usecs)
{
    size_t l = 0;
    ssize_t t;
    u8 *buf;
    byte_array *n;
    struct timeval *tv;
    u32 beginning, present;
    fd_set set;
    int ret;
    
    n = (byte_array *) xcalloc(1, sizeof(byte_array));
    if (secs == -1 && usecs == -1) {
        tv = NULL;
    } else {
        tv = (struct timeval *) alloca(sizeof(*tv));
        tv->tv_sec = secs;
        tv->tv_usec = usecs;
    }

    FD_ZERO(&set);
    FD_SET(s->fd, &set);

    beginning = now();
    errno = 0;
    while ((ret = select(s->fd + 1, &set, NULL, NULL, tv)) < 0) {
        if (errno == EINTR) {
            /* This is the only valid reason that select can fail afaik.  */
            /* This is imprecise, but select's resolution isn't perfect, so
               it's not a big loss.  */
            present = now();
            tv->tv_sec = secs - (beginning - present);
            if (tv->tv_sec < 0) {
                tv->tv_sec = 0;
                printf("Clock skew detected.\n");
            }
            tv->tv_usec = usecs;
            continue;
        } else {
            perror("select failed");
            abort();
        }
    }

    if (ret == 0) {
        goto done;
    }

    if (s->type == SOCKET_STREAM) {
        if (maxlen > 0) {
            buf = (u8 *) alloca(maxlen * sizeof(u8));
            t = recv(s->fd, buf, maxlen, 0);
            if (t == 0) {
                byte_array_destroy(n);
                goto done;
            }
            byte_array_cats(n, buf, t);
        } else {
            /* XXX: 512 is a magic number.  */
            buf = (u8 *) alloca(512 * sizeof(u8));
            do {
                t = recv(s->fd, buf, 512, 0);
                byte_array_cats(n, buf, t);
            } while (t > 0);
        }  
    } else if (s->type == SOCKET_DATAGRAM) {
        size_t f;
        if (s->domain == SOCKET_INTERNET) {
            f = sizeof(struct sockaddr_in);
        } else if (s->domain == SOCKET_LOCAL) {
            f = sizeof(struct sockaddr_un);
        } else {
            abort();
        }
        if (maxlen == 0 || maxlen > 512)
            maxlen = 512;
        buf = (u8 *) alloca(maxlen * sizeof(u8));
        t = recvfrom(s->fd, buf, maxlen, 0, (sa *) &s->socket_address.dummy,
                     &f);
        if (t == -1) {
            perror("recvfrom failed");
            byte_array_destroy(n);
            return 0;
        }
        byte_array_cats(n, buf, t);
    } else if (s->type == SOCKET_SEQPACKET) {
        ssize_t len;
        socket_set_minimum(s, 4);
        buf = (u8 *) alloca(4 * sizeof(u8));
        t = recv(s->fd, buf, 4, 0);
        if (t == 0) {
            byte_array_destroy(n);
            goto done;
        }
        len = ntohl(((u32 *) buf)[0]);
        socket_set_minimum(s, len);
        buf = (u8 *) alloca(len * sizeof(u8));
        t = recv(s->fd, buf, len, 0);
        if (t != len) {
            byte_array_destroy(n);
            goto done;
        }
        printf_on_debug("read '%s'\n", buf);
        byte_array_cats(n, buf, len);
    } else {
        abort();
    }
 
    byte_array_cat(data, n);
    byte_array_destroy(n);
    // free(buf);
    l = t;

done:
    return l;
}

/* Send data through a socket.  The semantics are, in this case, obvious.  */

size_t
socket_send(socket_t *s, byte_array *ba)
{
    return socket_send_timeout(s, ba, -1, -1);
}

size_t 
socket_send_poll(socket_t *s, byte_array *ba)
{
    return socket_send_timeout(s, ba, 0, 0);
}

size_t
socket_send_timeout(socket_t *s, byte_array *data, int secs, int usecs)
{
    size_t l = 0;
    ssize_t ll;
    struct timeval *tv;
    u32 beginning, present;
    fd_set set;
    int ret;

    byte_array *seq_ba = (byte_array *) 0x123;

    if (secs == -1 && usecs == -1) {
        tv = NULL;
    } else {
        tv = (struct timeval *) alloca(sizeof(*tv));
        tv->tv_sec = secs;
        tv->tv_usec = usecs;
    }

    FD_ZERO(&set);
    FD_SET(s->fd, &set);

    beginning = now();
    errno = 0;
    while ((ret = select(s->fd + 1, NULL, &set, NULL, tv)) <= 0) {
        if (errno == EINTR) {
            present = now();
            tv->tv_sec = secs - (beginning - present);
            if (tv->tv_sec < 0) {
                tv->tv_sec = 0;
                printf("Clock skew detected.\n");
            }
            tv->tv_usec = usecs;
            continue;
        } else {
            perror("select failed");
            abort();
        }
    }

    if (s->type == SOCKET_STREAM) {
        ll = send(s->fd, data->data, data->len, 0);
    } else if (s->type == SOCKET_DATAGRAM) {
        /* Goddamn!  sendto works fine in Linux but not in BSD.  The reason (as
           I've managed to deduce): specifying a peer that's already been
           specified - as it was in socket_sender via connect - is redundant
           and therefore (?) bad.  Hm.  I guess it's better to fail fast than
           to mysteriously fail and work.  Yet recvfrom as used above is OK.

           Goddamn!  But the above only holds if you're initiating the udp
           "connection".  If you're accepting, then this is incorrect.  

           Arg!  
           
           The newly redeployed type field is an attempt to compensate.  
           Transport is a more fitting name anyway.  */

        size_t n;
        
        if (s->domain == SOCKET_INTERNET) {
            n = sizeof(struct sockaddr_in);
        } else if (s->domain == SOCKET_LOCAL) {
            n = sizeof(struct sockaddr_un);
        } else {
            abort();
        }

        if (s->job == SOCKET_LISTENER) {
            ll = sendto(s->fd, data->data, data->len, 0, 
                        (sa *) &s->socket_address.dummy, n);
        } else if (s->job == SOCKET_SENDER) {
            ll = send(s->fd, data->data, data->len, 0);
        } else if (s->job == SOCKET_CLIENT) {
            ll = sendto(s->fd, data->data, data->len, 0,
                        (sa *) &s->socket_address.dummy, n);
        } else {
            abort();
        }

        if (ll == -1) {
            perror("sendto failed");
            abort();
        } else if (l > data->len) {
            perror("wtf?  sendto returned something larger than len");
            abort();
        }
    } else if (s->type == SOCKET_SEQPACKET) {
        // byte_array *t;
        static int foo = 1;
        
#ifdef WITH_DMALLOC
        int dmv_ret = 0;
        // dmv_ret = dmalloc_verify(seq_ba);
#endif


        // printf("been here %d times now\n", foo); ++foo;
        /* The line immediately below causes a failure when some forms
           of malloc checking (dmalloc, glibc's MALLOC_CHECK_) but not
           others (efence).  I do not know why.  */
        seq_ba = (byte_array *) xcalloc(1, sizeof(byte_array));
        byte_array_catu32(seq_ba, htonl(data->len));
        printf_on_debug("length is %d\n", seq_ba->len);
        // printf_on_debug("appending this: '%s'\n", data->data);
        byte_array_cat(seq_ba, data);
        ll = send(s->fd, seq_ba->data, seq_ba->len, 0);
        byte_array_destroy(seq_ba);
    } else {
        abort();
    }
 
    return ll;
}

void
socket_set_minimum(socket_t *s, size_t min)
{
    int ret;
    
    ret = setsockopt(s->fd, SOL_SOCKET, SO_RCVLOWAT, &min, sizeof(min));
    if (ret == -1) {
        perror("setsockopt failed");
        abort();
    }
}

int
socket_seqpacket_handshake(socket_t *s)
{
    int retval;
    const u8 *msg1 = "SOCKET_SEQPACKET";
    const u8 *msg2 = "SOCKET_SEQPACKET ACCEPTED";
    const u8 *msg3 = "OK";
    int state;
    u8 *buf;
    size_t buflen;
    int ret;
    
    assert(s->type == SOCKET_SEQPACKET);
    if (s->job == SOCKET_SENDER)
        state = 0;
    else if (s->job == SOCKET_CLIENT)
        state = 10;
    else
        abort();

    buflen = strlen(msg2);
    buf = (u8 *) alloca(buflen + 1);
    
    while (1) {
        switch (state) {
        case 0:
            ret = handshake_send(msg1, s);
            if (ret == 0)
                state = 999;
            else
                state = 1;
            break;
        case 1:
            ret = handshake_recv(buf, msg2, buflen, s);
            if (ret == 0)
                state = 999;
            else
                state = 2;
            break;
        case 2:
            ret = handshake_send(msg3, s);
            if (ret == 0)
                state = 999;
            else
                state = 3;
            break;
        case 3:
            ret = handshake_recv(buf, msg3, strlen(msg3), s);
            if (ret == 0)
                state = 999;
            else
                state = 4;
            break;
        case 4:
            /* I think we're all set.  */
            state = 1000;
            break;
        case 10:
            ret = handshake_recv(buf, msg1, strlen(msg1), s);
            if (ret == 0)
                state = 999;
            else
                state = 11;
            break;
        case 11:
            ret = handshake_send(msg2, s);
            if (ret == 0)
                state = 999;
            else
                state = 12;
            break;
        case 12:
            ret = handshake_recv(buf, msg3, strlen(msg3), s);
            if (ret == 0)
                state = 999;
            else
                state = 13;
            break;
        case 13:
            ret = handshake_send(msg3, s);
            if (ret == 0)
                state = 999;
            else
                state = 14;
            break;
        case 14:
            /* I think we're all set.  */
            state = 1000;
            break;
        case 999:
            /* Cleanup on failure.  */
            /* XXX: actually do this.  */
            goto done;
            break;
        case 1000:
            retval = 1;
            /* Cleanup after success.  */
            /* XXX: actually do this.  */
            goto done;
            break;
        default:
            fprintf(stderr, "screwed up state\n");
            abort();
        }
    }

done:
    return retval;
}

int
handshake_send(const char *msg, socket_t *s)
{
    int retval;
    int ret;
    fd_set wset;
    struct timeval tv;
    ssize_t ll;
    ssize_t mlen;

    

    mlen = strlen(msg);
    FD_ZERO(&wset);
    FD_SET(s->fd, &wset);
    tv.tv_sec = 2000;
    tv.tv_usec = 0;
    while ((ret = select(s->fd + 1, NULL, &wset, NULL, &tv)) <= 0) {
        if (errno == EINTR)
            continue;
        else {
            perror("select failed");
            abort();
        }
    }
    ll = send(s->fd, msg, mlen, 0);
    if (ll < mlen)
        retval = 0;
    else
        retval = 1;

    return retval;
}

int
handshake_recv(char *msg, const char *expected_text, 
               size_t expected_len, socket_t *s)
{
    int retval;
    int ret;
    fd_set rset;
    struct timeval tv;
    ssize_t ll;
    
    socket_set_minimum(s, expected_len);
    
    FD_ZERO(&rset);
    FD_SET(s->fd, &rset);
    tv.tv_sec = 2000;
    tv.tv_usec = 0;
    memset(msg, '\0', expected_len);
    while ((ret = select(s->fd + 1, &rset, NULL, NULL, &tv)) <= 0) {
        if (errno == EINTR)
            continue;
        else {
            perror("select failed");
            abort();
        }
    }
    if (ret == 0) {
        retval = 0;
        goto done;
    }        
    ll = recv(s->fd, msg, expected_len, 0);
    /* This shouldn't be necessary due to the socket option set above.  */
    /* if (ll != expected_len)
       retval = 0;  */
    if (strncmp(msg, expected_text, expected_len) == 0)
        retval = 1;
    else
        retval = 0;
done:
    return retval;
}


void
socket_set_attribute_blocking(socket_t *s, enum socket_blocking sab)
{   
    s->blocking_attribute = sab;    
}
