/* 
 * Part of the pthread_utils package.  
 *
 * Copyright (c) 2001 Michael Wolf.  All rights reserved.  
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* $Header: /cvsroot/moodns/moodns/src/wpthread.c,v 1.2 2001/12/21 13:28:34 mawolf Exp $  */
#ifndef _REENTRANT
#define _REENTRANT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h> /* for strerror.  */

#include <pthread.h>
#include "wpthread.h"

static void _print_error(const char *, int);

void
wpthread_cond_init(pthread_cond_t *cond, const pthread_condattr_t *condattr)
{
    int ret;

    ret = pthread_cond_init(cond, condattr);
    if (ret != 0) {
        _print_error("pthread_cond_init() failed", ret);
        abort();
    }
}

void
wpthread_cond_signal(pthread_cond_t *cond)
{
    int ret;
    
    ret = pthread_cond_signal(cond);
    if (ret != 0) {
        _print_error("pthread_cond_signal() failed", ret);
        abort();
    }
}

void 
wpthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *mutexaddr)
{
    int ret;

    ret = pthread_mutex_init(mutex, mutexaddr);
    if (ret != 0) {
        _print_error("pthread_mutex_init() failed", ret);
        abort();
    }
}

void
wpthread_create(pthread_t *thread, const pthread_attr_t *attr, 
                void *(*func)(void *), void *aux)
{
    int ret;

    ret = pthread_create(thread, attr, func, aux);
    if (ret != 0) {
        _print_error("pthread_create() failed", ret);
        abort();
    }
}

/* 
 * I'm not so sure about this one.  I think whether it works might depend on
 * how pthread_t is composed, which is implementation defined.  I think it 
 * should work in any case, but it does highlight what seems like a blemish,
 * or at least an unusual inconsistency, in the way parameters are passed.
 */
void
wpthread_join(pthread_t thread, void **aux) 
{
    int ret;

    ret = pthread_join(thread, aux);
    if (ret != 0) {
        _print_error("pthread_join() failed", ret);
        abort();
    }
}

void
wpthread_detach(pthread_t thread)
{
    int ret;

    ret = pthread_detach(thread);
    if (ret != 0) {
        _print_error("pthread_detach() failed", ret);
        abort();
    }
}

void 
wpthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
    int ret;

    ret = pthread_cond_wait(cond, mutex);
    if (ret != 0) {
        _print_error("pthread_cond_wait() failed", ret);
        abort();
    }
}

void
wpthread_mutex_lock(pthread_mutex_t *mutex)
{
    int ret;

    ret = pthread_mutex_lock(mutex);
    if (ret != 0) {
        _print_error("pthread_mutex_lock() failed", ret);
        abort();
    }
}

void
wpthread_mutex_unlock(pthread_mutex_t *mutex)
{
    int ret;

    ret = pthread_mutex_unlock(mutex);
    if (ret != 0) {
        _print_error("pthread_mutex_unlock() failed", ret);
        abort();
    }
}

void
wpthread_cond_broadcast(pthread_cond_t *cond)
{
    int ret;
    
    ret = pthread_cond_broadcast(cond);
    if (ret != 0) {
        _print_error("pthread_cond_broadcast() failed", ret);
        abort();
    }
}

void
wpthread_once(pthread_once_t *eleven, void (*func)(void))
{
    int ret;

    ret = pthread_once(eleven, func);
    if (ret != 0) {
        _print_error("pthread_once() failed", ret);
        abort();
    }
}

pthread_t
wpthread_self(void)
{
    pthread_t retval;
    
    retval = pthread_self();
    /* Can this fail?  

    Answer: It can, at least in some cases.  For example, on Solaris, 
    applications which depend on pthreads will build without being linked 
    against libpthread.  In that case, stubs which always fail are called 
    instead.  It's not clear that we can do anything about this here.  However,
    if we're using pthread_self(), we're probably using other pthread functions
    too, and those *will* fail.  */
   
    return (retval);
}

int
wpthread_equal(pthread_t a, pthread_t b)
{
    /* Can this fail?  */
    return (pthread_equal(a, b));
}

int
wpthread_mutex_trylock(pthread_mutex_t *mutex)
{
    return (pthread_mutex_trylock(mutex));
}

void
_print_error(const char *message, int error)
{
    fprintf(stderr, "%s: %s\n", message, strerror(error));
}

